/*
 * CoreLogic.java
 *
 * Created on May 24, 2004, 3:17 PM
 */

package gov.bnl.gums;

import gov.bnl.gums.account.*;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.configuration.ConfigurationToolkit;
import gov.bnl.gums.groupToAccount.GroupToAccountMapping;
import gov.bnl.gums.hostToGroup.HostToGroupMapping;
import gov.bnl.gums.userGroup.UserGroup;
import gov.bnl.gums.userGroup.VOMSUserGroup;
import gov.bnl.gums.util.StringUtil;

import java.io.FileInputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.*;

import org.apache.log4j.Logger;

/** 
 * The main interface of the core logic for GUMS, allows to update the
 * groups, generate the gridMapfiles and to map a single user.
 *
 * @author Gabriele Carcassi, Jay Packard
 */
public class CoreLogic {
    //private Logger log = Logger.getLogger(CoreLogic.class); // only use this log for particularly tricky aspects of this class - otherwise log within lower level classes
    private Logger gumsAdminLog = Logger.getLogger(GUMS.gumsAdminLogName);
    Configuration conf = null;

    /**
     * Creates a Core Logic object to use the singleton GUMS object to get the configuration
     * 
     * @param configuration
     */
    public CoreLogic() {
    }
    
    /**
     * Creates a Core Logic object for a certain configuration (usually for testing)
     * 
     * @param configuration
     */
    public CoreLogic(Configuration conf) {
    	this.conf = conf;
    }
    
    public void addAccountRange(String accountPoolMapperName, String range) throws Exception {
		AccountMapper accountMapper = getConfiguration().getAccountMapper(accountPoolMapperName);
		if (accountMapper==null)
			throw new RuntimeException("Account Mapper '" + accountPoolMapperName + "' does not exist");
		if (!(accountMapper instanceof AccountPoolMapper))
			throw new RuntimeException("Account Mapper '" + accountPoolMapperName + "' is not an account pool mapper");
		AccountPoolMapper accountPoolMapper = (AccountPoolMapper)accountMapper;

		String firstAccount = range.substring(0, range.indexOf('-'));
		String lastAccountN = range.substring(range.indexOf('-') + 1);
		String firstAccountN = firstAccount.substring(firstAccount.length() - lastAccountN.length());
		//String accountBase = firstAccount.substring(0, firstAccount.length() - lastAccountN.length());
		int nFirstAccount = Integer.parseInt(firstAccountN);
		int nLastAccount = Integer.parseInt(lastAccountN);

		StringBuffer last = new StringBuffer(firstAccount);
		String nLastAccountString = Integer.toString(nLastAccount);
		last.replace(firstAccount.length() - nLastAccountString.length(), firstAccount.length(), nLastAccountString);

		StringBuffer buf = new StringBuffer(firstAccount);
		int len = firstAccount.length();
		List<Mapping> mappings = ((AccountPoolMapper)accountPoolMapper).getMappings();
		for (int account = nFirstAccount; account <= nLastAccount; account++) {
			String nAccount = Integer.toString(account);
			buf.replace(len - nAccount.length(), len, nAccount);
			if (mappings.contains(new SiteUser(buf.toString()))) {
				String message = "One or more accounts already exist for '"+accountPoolMapperName+" "+range+"'. None were added.";
				gumsAdminLog.warn(message);
				throw new RuntimeException(message);
			}
		}
		buf = new StringBuffer(firstAccount);
		for (int account = nFirstAccount; account <= nLastAccount; account++) {
			String nAccount = Integer.toString(account);
			buf.replace(len - nAccount.length(), len, nAccount);
			accountPoolMapper.createMapping(null, new SiteUser(buf.toString()));
		}
    }
    
    public boolean isUserBanned(GridUser user) throws Exception {
    	return getConfiguration().isUserBanned(user);
    }
    
     /**
     * @param hostname
     * @return
     */
    public String generateOsgUserVoMap(String hostname) throws Exception {
        Configuration conf = getConfiguration();
        StringBuffer osgMapBuffer = new StringBuffer("");
        HostToGroupMapping host2GroupMapper = hostToGroupMapping(conf, hostname);
        if (host2GroupMapper == null) {
        	String message = "Cannot generate osg user VO map for host '" + hostname + "' - it is not defined in any host to group mapping.";
            gumsAdminLog.warn(message);
        	throw new RuntimeException(message);
        }
        
        // Create header
        osgMapBuffer.append("#User-VO map\n");
        osgMapBuffer.append("# #comment line, format of each regular line line: account VO\n");
        osgMapBuffer.append("# Next 2 lines with VO names, same order, all lowercase, with case (lines starting with #voi, #VOc)\n");
        String voi = "#voi";
        String voc = "#VOc";
        Iterator<String> iter = getGroupToAccountMappings(host2GroupMapper).iterator();
        while (iter.hasNext()) {
            GroupToAccountMapping gMap = conf.getGroupToAccountMapping( iter.next() );
            Collection<String> userGroups = getUserGroups(gMap);
            Iterator<String> userGroupIt = userGroups.iterator();
            while (userGroupIt.hasNext()) {
            	UserGroup userGroup = conf.getUserGroup( userGroupIt.next() );
	            if (userGroup.getMembers().size()!=0 && !gMap.getAccountingVoSubgroup().equals("") && !gMap.getAccountingVo().equals("")) {
	                voi = voi + " " + gMap.getAccountingVoSubgroup();
	                voc = voc + " " + gMap.getAccountingVo();
	            }
            }
        }
        osgMapBuffer.append(voi);
        osgMapBuffer.append("\n");
        osgMapBuffer.append(voc);
        osgMapBuffer.append("\n");
        
        // Loop through group to account mappings
        List<String> accountsInMap = new ArrayList<String>();
        iter = getGroupToAccountMappings(host2GroupMapper).iterator();
        while (iter.hasNext()) {
            GroupToAccountMapping gMap = (GroupToAccountMapping) conf.getGroupToAccountMapping( (String)iter.next() );
            if (!gMap.getAccountingVoSubgroup().equals("") && !gMap.getAccountingVo().equals("")) {
	            Collection<String> userGroups = getUserGroups(gMap);
            	Collection<String> accountMappers = getAccountMappers(gMap);
	            Iterator<String> userGroupIt = userGroups.iterator();
	            while (userGroupIt.hasNext()) {
	            	UserGroup userGroup = conf.getUserGroup( userGroupIt.next() );
		            Set<GridUser> users = new TreeSet<GridUser>(retrieveUserComparatorByDN());
		            users.addAll(userGroup.getMembers());
		            osgMapBuffer.append("#---- accounts for vo: " + userGroup.getName() + " ----#\n");          
		            Iterator<GridUser> memberIter = users.iterator();
		            while (memberIter.hasNext()) {
		                GridUser user = memberIter.next();
		                if (isUserBanned(user))
		                	continue;
	                    Iterator<String> accountMapperIt = accountMappers.iterator();
	                    while (accountMapperIt.hasNext()) {
	                    	AccountMapper accountMapper = conf.getAccountMapper( accountMapperIt.next() );
			                SiteUser siteUser = accountMapper.mapDn(user.getDn(), false);
			                if (siteUser!=null && siteUser.getAccount()!=null && !accountsInMap.contains(siteUser.getAccount())) {
			                	osgMapBuffer.append(siteUser.getAccount());
			                	osgMapBuffer.append(' ');
			                	osgMapBuffer.append(gMap.getAccountingVoSubgroup());
			                	osgMapBuffer.append("\n");
			                    accountsInMap.add(siteUser.getAccount());
			                }
	                    }
		            }
	            }
            }
        }
        
        String finalOsgMap = osgMapBuffer.toString();
        return finalOsgMap;
    }
    

    /**
     * Generates a fqan to account mapping for a given host and prints it to out.
     * 
     * @param hostname
     * @return
     */
    public String generateFqanMapfile(String hostname, boolean getAssignments) throws Exception {
        String mapfile;
        mapfile = generateFqanMapfileImpl(hostname, getAssignments);
        return mapfile;
    }    
    
    /**
     * Generates a grid mapfile for a given host and prints it to out.
     * 
     * @param hostname
     * @param includeFQAN
     * @return
     */
    public String generateGridMapfile(String hostname, boolean createNewMappings, boolean includeFQAN, boolean includeEmail) throws Exception {
        String mapfile;
        mapfile = generateGridMapfileImpl(hostname, createNewMappings, includeFQAN, includeEmail);
        return mapfile;
    }
    
    /**
     * Get a list of out of date FQANS (because of inability to contact VOMS server for example)
     * 
     * @return
     */
    public String getOutOfDateFqans(int hoursSinceLastUpdate) throws Exception {
    	StringBuffer sb = new StringBuffer();
    	for (UserGroup userGroup: getConfiguration().getUserGroups().values()) {
    		if (userGroup instanceof VOMSUserGroup) {
    			VOMSUserGroup vomsUserGroup = (VOMSUserGroup)userGroup;
    			Calendar cal = Calendar.getInstance();
				cal.add(Calendar.HOUR, -hoursSinceLastUpdate);
    			if (vomsUserGroup.getLastUpdate().before(cal.getTime()))
    				sb.append(userGroup.getName() + " " + userGroup.getLastUpdate());
    		}
    	}
    	return sb.toString();
    }
    
    /**
     * Maps a user to a local account for a given host to a grid identity.
     * 
     * @param hostname
     * @param user
     * @return
     */
    public SiteUser map(String hostname, GridUser user, boolean checkBannedList) throws Exception {
        if (checkBannedList && isUserBanned(user))
        	return null;
        return mapImpl(hostname, user);
    }
    
    /**
     * Map a local account to a list of grid identity
     * 
     * @param accountName
     * @return
     */
    public String mapAccount(String accountName) throws Exception {
    	TreeSet<String> dns = new TreeSet<String>();
        Configuration conf = getConfiguration();
        Iterator<GroupToAccountMapping> g2AMappingsIt = conf.getGroupToAccountMappings().values().iterator();
        while (g2AMappingsIt.hasNext()) {
            GroupToAccountMapping g2AMapping = g2AMappingsIt.next();
            Collection<String> userGroups = getUserGroups(g2AMapping);
            Iterator<String> userGroupsIt = userGroups.iterator();
            while (userGroupsIt.hasNext()) {
            	UserGroup userGroup = conf.getUserGroup( userGroupsIt.next() );
            	Set<GridUser> users = userGroup.getMembers();
            	Iterator<GridUser> usersIt = users.iterator();
            	while (usersIt.hasNext()) {
            		GridUser user = usersIt.next();
            		if (isUserBanned(user))
                    	continue;
                	Collection<String> accountMappers = getAccountMappers(g2AMapping);
                    Iterator<String> accountMappersIt = accountMappers.iterator();
                    while (accountMappersIt.hasNext()) {
                    	AccountMapper accountMapper = conf.getAccountMapper( accountMappersIt.next() );
                		if (accountName.equals(accountMapper.mapDn(user.getDn(), false)) && !dns.contains(user.getDn()))
                        	dns.add(user.getDn());
                    }            		
            	}
            }
        }
        
        if (dns.size()==0) {
            gumsAdminLog.warn("Cannot map account '" + accountName + "' to any DN.");
        }
        
        StringBuffer output = new StringBuffer();
        Iterator<String> it = dns.iterator();
        while(it.hasNext()) {
        	output.append( it.next() );
        	output.append("\n");
        }

		gumsAdminLog.info("Mapped the account '" + accountName + "' to '" + output.toString() + "'");
        return output.toString();
    }
    
    /**
     * @param hostname
     * @return
     */
    private String generateFqanMapfileForH2GMapping(HostToGroupMapping hToGMapping, boolean getAssignments) throws Exception {   
        StringBuffer fqanMapfileBuffer = new StringBuffer("");
      	
        if (hToGMapping != null) {
           	Configuration conf = hToGMapping.getConfiguration();
	       	Iterator<String> iter = getGroupToAccountMappings(hToGMapping).iterator();
	        while (iter.hasNext()) {
	            GroupToAccountMapping gMap = (GroupToAccountMapping)conf.getGroupToAccountMapping( iter.next() );
	            fqanMapfileBuffer.append(generateFqanMapfileForG2AMapping(conf, gMap, getAssignments));
	        }
        }
        else {
        	Configuration conf = getConfiguration();
	       	Iterator<GroupToAccountMapping> iter = conf.getGroupToAccountMappings().values().iterator();
	        while (iter.hasNext()) {
	            GroupToAccountMapping gMap = iter.next();
	            fqanMapfileBuffer.append(generateFqanMapfileForG2AMapping(conf, gMap, getAssignments));
	        }
        }
        
        return fqanMapfileBuffer.toString();
    }    
    
    private String generateFqanMapfileForG2AMapping(Configuration conf, GroupToAccountMapping gMap, boolean getAssignments) {
        StringBuffer fqanMapfileBuffer = new StringBuffer("");
        Collection<String> userGroups = getUserGroups(gMap);
        Iterator<String> userGroupIt = userGroups.iterator();
        while (userGroupIt.hasNext()) {
        	UserGroup userGroup = conf.getUserGroup( userGroupIt.next() );
        	if (userGroup instanceof VOMSUserGroup) {
        		VOMSUserGroup vomsUserGroup = (VOMSUserGroup)userGroup;
        		Collection<String> accountMappers = getAccountMappers(gMap);
                Iterator<String> accountMapperIt = accountMappers.iterator();
                while (accountMapperIt.hasNext()) {
                	AccountMapper accountMapper = conf.getAccountMapper( accountMapperIt.next() );
                	if (accountMapper instanceof GroupAccountMapper || accountMapper instanceof AccountPoolMapper) {
                		String fqan = vomsUserGroup.getVoGroup() + (vomsUserGroup.getRole().equals("") ? "" : "/Role=" + vomsUserGroup.getRole());
                    	String account = null;
                       	if (accountMapper instanceof GroupAccountMapper) {
                    		GroupAccountMapper groupPoolMapper = (GroupAccountMapper)accountMapper;
                    		account = groupPoolMapper.getAccountName();
                    	}
                    	else if (accountMapper instanceof AccountPoolMapper){
                    		if (getAssignments)
                    			account = ((AccountPoolMapper)accountMapper).getAssignments();
                    		else {
	                    		String accountPool = ((AccountPoolMapper)accountMapper).getAccountPool();
	                    		account = accountPool + "0-"+accountPool+"0(0/0)";
                    		}
                    	}
                       	if (account != null) {
                       		fqanMapfileBuffer.append('"');
                       		fqanMapfileBuffer.append( fqan );
                       		fqanMapfileBuffer.append('"');
                       		fqanMapfileBuffer.append(' ');
                       		fqanMapfileBuffer.append( account );
                       		fqanMapfileBuffer.append("\n");
                       		break;
                       	}
                	}
                }
        	}
        }
        return fqanMapfileBuffer.toString();
    }
    
    private String generateGridMapfile(HostToGroupMapping hToGMapping, boolean createNewMappings, boolean includeFQAN, boolean includeEmail) throws Exception {
    	Configuration conf = hToGMapping.getConfiguration();
        Iterator<String> iter = getGroupToAccountMappings(hToGMapping).iterator();
        TreeSet<String> usersInMap = new TreeSet<String>();
        
        StringBuffer gridMapfileBuffer = new StringBuffer("");
        while (iter.hasNext()) {
            GroupToAccountMapping gMap = (GroupToAccountMapping) conf.getGroupToAccountMapping( (String)iter.next() );
            Collection<String> userGroups = getUserGroups(gMap);
            Iterator<String> userGroupIt = userGroups.iterator();
            while (userGroupIt.hasNext()) {
            	UserGroup userGroup = conf.getUserGroup( userGroupIt.next() );
	            Set<GridUser> members = new TreeSet<GridUser>(retrieveUserComparatorByDN());
	            members.addAll(userGroup.getMembers());
	            gridMapfileBuffer.append("#---- members of vo: " + userGroup.getName() + " ----#\n");
	            Iterator<GridUser> memberIter = members.iterator();
	            if (includeFQAN) {
		            while (memberIter.hasNext()) {
		                GridUser user = memberIter.next();
		                if (isUserBanned(user))
	                    	continue;
	                    String fqan = user.getFqan()!=null ? user.getFqan() : "";
	                    String email = user.getEmail()!=null?user.getEmail().toString():"";
		                if ( !usersInMap.contains(user.getDn() + fqan) ) {
		                	Collection<String> accountMappers = getAccountMappers(gMap);
		                    Iterator<String> accountMappersIt = accountMappers.iterator();
		                    while (accountMappersIt.hasNext()) {
		                    	AccountMapper accountMapper = conf.getAccountMapper( accountMappersIt.next() );
			                	SiteUser siteUser = accountMapper.mapDn(user.getDn(), true);
			                	if (siteUser!=null && siteUser.getAccount()!=null) {
			                        gridMapfileBuffer.append('"');
			                        gridMapfileBuffer.append( user.getDn() );
			                        gridMapfileBuffer.append('"' );
			                        gridMapfileBuffer.append(" \"");
			                        gridMapfileBuffer.append( fqan );
			                        gridMapfileBuffer.append('"');         	
			                        gridMapfileBuffer.append(' ');
			                        gridMapfileBuffer.append( siteUser.getAccount() );
			                        if (includeEmail) {
				                        gridMapfileBuffer.append(' ');
				                        gridMapfileBuffer.append( email );
			                        }
			                        gridMapfileBuffer.append("\n");
		                        	usersInMap.add( user.getDn() + fqan );
		                        	break;
			                    } else {
			                    	gumsAdminLog.debug("User " + user + " from group " + gMap.getUserGroups() + " can't be mapped.");
			                    }
		                    }
		            	}
		            }
	            }
	            else {
		            while (memberIter.hasNext()) {
		                GridUser user = (GridUser) memberIter.next();
		                if (isUserBanned(user))
	                    	continue;
		                String email = user.getEmail()!=null?user.getEmail().toString():"";
		                if (!usersInMap.contains(user.getDn()) && userGroup.isMember(new GridUser(user.getDn()))) {
		                	Collection<String> accountMappers = getAccountMappers(gMap);
		                    Iterator<String> accountMappersIt = accountMappers.iterator();
		                    while (accountMappersIt.hasNext()) {
		                    	AccountMapper accountMapper = conf.getAccountMapper( accountMappersIt.next() );
			                	SiteUser siteUser = accountMapper.mapDn(user.getDn(), true);
			                    if (siteUser!=null && siteUser.getAccount()!=null) {
			                        gridMapfileBuffer.append('"');
			                        gridMapfileBuffer.append( user.getDn() );
			                        gridMapfileBuffer.append('"' );
			                        gridMapfileBuffer.append(' ');
			                        gridMapfileBuffer.append( siteUser.getAccount() );
			                        if (includeEmail && email!=null) {
				                        gridMapfileBuffer.append(' ');
				                        gridMapfileBuffer.append( email );
			                        }
			                        gridMapfileBuffer.append("\n");
		                        	usersInMap.add(user.getDn());
		                        	break;
			                    } else {
			                    	gumsAdminLog.debug("User " + user + " from group " + gMap.getUserGroups() + " can't be mapped.");
			                    }
		                    }
		            	}
		            }	            	
	            }
            }
        }

        String finalGridmapfile = gridMapfileBuffer.toString();
        return finalGridmapfile;
    }
 
    private String generateFqanMapfileImpl(String hostname, boolean getAssignments) throws Exception {
        String mapfile = null;
        if (hostname != null) {
	    	Configuration conf = getConfiguration();
	        HostToGroupMapping host2GroupMapper = hostToGroupMapping(conf, hostname);
	        if (host2GroupMapper == null) {
	        	String message = "Cannot generate FQAN mapfile for host '" + hostname + "' - it is not defined in any host to group mapping.";
	            gumsAdminLog.warn(message);
	        	throw new RuntimeException(message);
	        }
	        mapfile = generateFqanMapfileForH2GMapping(host2GroupMapper, getAssignments);
        }
        else
        	mapfile = generateFqanMapfileForH2GMapping(null, getAssignments);
        return mapfile;
    }
    
    private String generateGridMapfileImpl(String hostname, boolean createNewMappings, boolean includeFqan, boolean includeEmail) throws Exception {
    	Configuration conf = getConfiguration();
        HostToGroupMapping host2GroupMapper = hostToGroupMapping(conf, hostname);
        if (host2GroupMapper == null) {
        	String message = "Cannot generate grid mapfile for host '" + hostname + "' - it is not defined in any host to group mapping.";
            gumsAdminLog.warn(message);
        	throw new RuntimeException(message);
        }
        String mapfile = generateGridMapfile(host2GroupMapper, createNewMappings, includeFqan, includeEmail);
        return mapfile;
    }
    
    private Configuration getConfiguration() throws Exception {
    	if (conf != null)
    		return conf;
    	else
    		return GUMS.getSingletonGums().getConfiguration();
    }
    
    private HostToGroupMapping hostToGroupMapping(Configuration conf, String hostname) {
    	Collection<HostToGroupMapping> hostToGroupMappers = conf.getHostToGroupMappings();
    	Iterator<HostToGroupMapping> it = hostToGroupMappers.iterator();
        while (it.hasNext()) {
            HostToGroupMapping hostToGroupMapper = it.next();
            if (hostToGroupMapper.isInGroup(hostname)) {
                return hostToGroupMapper;
            }
        }
        return null;
    }

    private SiteUser mapImpl(String hostname, GridUser user) throws Exception {
    	Configuration conf = getConfiguration();
        HostToGroupMapping host2GroupMapper = hostToGroupMapping(conf, hostname);
        if (host2GroupMapper == null) {
        	String message = "Cannot map user for host '" + hostname + "' - it is not defined in any host to group mapping.";
            gumsAdminLog.warn(message);
        	throw new RuntimeException(message);
        }
        Iterator<String> g2AMappingsIt = getGroupToAccountMappings(host2GroupMapper).iterator();
        while (g2AMappingsIt.hasNext()) {
            GroupToAccountMapping g2AMapping = conf.getGroupToAccountMapping( g2AMappingsIt.next() );
            Collection<String> userGroups = getUserGroups(g2AMapping);
            Iterator<String> userGroupsIt = userGroups.iterator();
            while (userGroupsIt.hasNext()) {
            	UserGroup userGroup = conf.getUserGroup( userGroupsIt.next() );
                if (userGroup.isMember(user)) {
                	Collection<String> accountMappers = getAccountMappers(g2AMapping);
                    Iterator<String> accountMappersIt = accountMappers.iterator();
                    while (accountMappersIt.hasNext()) {
                    	AccountMapper accountMapper = conf.getAccountMapper( accountMappersIt.next() );
                        SiteUser siteUser = accountMapper.mapDn(user.getDn(), true);
                        if (siteUser!=null && siteUser.getAccount()!=null)
                            return siteUser;
                    }
                }
            }
        }
        return null;
    }

	public void removeAccountRange(String accountPoolMapperName, String range) throws Exception {
		AccountMapper accountMapper = getConfiguration().getAccountMapper(accountPoolMapperName);
		if (accountMapper==null)
			throw new RuntimeException("Account Mapper '" + accountPoolMapperName + "' does not exist");
		if (!(accountMapper instanceof AccountPoolMapper))
			throw new RuntimeException("Account Mapper '" + accountPoolMapperName + "' is not an account pool mapper");
		AccountPoolMapper accountPoolMapper = (AccountPoolMapper)accountMapper;

		String firstAccount = range.substring(0, range.indexOf('-'));
		String lastAccountN = range.substring(range.indexOf('-') + 1);
		String firstAccountN = firstAccount.substring(firstAccount.length() - lastAccountN.length());
		//String accountBase = firstAccount.substring(0, firstAccount.length() - lastAccountN.length());
		int nFirstAccount = Integer.parseInt(firstAccountN);
		int nLastAccount = Integer.parseInt(lastAccountN);

		StringBuffer last = new StringBuffer(firstAccount);
		String nLastAccountString = Integer.toString(nLastAccount);
		last.replace(firstAccount.length() - nLastAccountString.length(), firstAccount.length(), nLastAccountString);

		StringBuffer buf = new StringBuffer(firstAccount);
		int len = firstAccount.length();
		for (int account = nFirstAccount; account <= nLastAccount; account++) {
			String nAccount = Integer.toString(account);
			buf.replace(len - nAccount.length(), len, nAccount);
			accountPoolMapper.removeMapping(new SiteUser(buf.toString()));
		}
	}
    
    private Comparator<GridUser> retrieveUserComparatorByDN() {
        return new Comparator<GridUser>() {
            public int compare(GridUser u1, GridUser u2) {
                GridUser user1 = u1;
                GridUser user2 = u2;
                return user1.compareDn(user2);
            }
        };
    }
    
	public void unassignAccountRange(String accountPoolMapperName, String range) throws Exception {
		AccountMapper accountMapper = getConfiguration().getAccountMapper(accountPoolMapperName);
		if (accountMapper==null)
			throw new RuntimeException("Account Mapper '" + accountPoolMapperName + "' does not exist");
		if (!(accountMapper instanceof AccountPoolMapper))
			throw new RuntimeException("Account Mapper '" + accountPoolMapperName + "' is not an account pool mapper");
		AccountPoolMapper accountPoolMapper = (AccountPoolMapper)accountMapper;
		
		String firstAccount = range.substring(0, range.indexOf('-'));
		String lastAccountN = range.substring(range.indexOf('-') + 1);
		String firstAccountN = firstAccount.substring(firstAccount.length() - lastAccountN.length());
		//String accountBase = firstAccount.substring(0, firstAccount.length() - lastAccountN.length());
		int nFirstAccount = Integer.parseInt(firstAccountN);
		int nLastAccount = Integer.parseInt(lastAccountN);

		StringBuffer last = new StringBuffer(firstAccount);
		String nLastAccountString = Integer.toString(nLastAccount);
		last.replace(firstAccount.length() - nLastAccountString.length(), firstAccount.length(), nLastAccountString);

		StringBuffer buf = new StringBuffer(firstAccount);
		int len = firstAccount.length();
		for (int account = nFirstAccount; account <= nLastAccount; account++) {
			String nAccount = Integer.toString(account);
			buf.replace(len - nAccount.length(), len, nAccount);
			accountPoolMapper.createMapping(null, new SiteUser(buf.toString()));
		}
	}

    
	public static void main(String args[]) {
		if (args.length < 2) {
			System.out.println("Must specify arguments: {configuration path} {command} {argument1} ...");
			return;
		}
		
        // Disable system error output since getBuilderFactory is outputting a bunch of information messages to it
    	PrintStream errStreamOriginal = System.err;
    	System.setErr(new PrintStream(new OutputStream() {
    		public void write(int b) {
    		}
    	}));
    	PrintStream outStreamOriginal = System.out;
    	System.setOut(new PrintStream(new OutputStream() {
    		public void write(int b) {
    		}
    	}));
		
		String configurationPath = args[0];
		String command = args[1];
		String map = "";
		try {
			FileInputStream fileInputStream = new FileInputStream(configurationPath);
			try {
				StringBuffer configBuffer = new StringBuffer();
				int ch;
				while ((ch = fileInputStream.read()) != -1)
					configBuffer.append((char)ch);
				Configuration conf = ConfigurationToolkit.parseConfiguration(configBuffer.toString(), false);
				CoreLogic coreLogic = new CoreLogic(conf);
				if (command.equals("generateFqanMapfile")) {
					if (args.length == 3)
						map = coreLogic.generateFqanMapfile(args[2], false);
					else
						map = coreLogic.generateFqanMapfileImpl(null, false);
				}
				else
					System.out.println("Command '" + command + "' not supported");
			} finally {
				fileInputStream.close();
			}
		} catch (Exception e) {
			System.setErr(errStreamOriginal);
			System.setOut(outStreamOriginal);
		 	System.out.println(e.getMessage());
		 	System.exit(1);
		} 
		
		System.setErr(errStreamOriginal);
		System.setOut(outStreamOriginal);
		
		System.out.println(map);
	}
	
	protected List<String> getUserGroups(GroupToAccountMapping g2aMapping) {
		return StringUtil.toList(g2aMapping.getUserGroups());
	}
	
	protected List<String> getAccountMappers(GroupToAccountMapping g2aMapping) {
		return StringUtil.toList(g2aMapping.getAccountMappers());
	}
	
	protected List<String> getGroupToAccountMappings(HostToGroupMapping h2gMapping) {
		return StringUtil.toList(h2gMapping.getGroupToAccountMappings());
	}
}
