/*
 * GUMS.java
 *
 * Created on June 3, 2004, 10:39 AM
 */

package gov.bnl.gums;

import gov.bnl.gums.admin.CertCache;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.configuration.ConfigurationToolkit;
import gov.bnl.gums.configuration.Version;
import gov.bnl.gums.persistence.HibernatePersistenceEngine;
import gov.bnl.gums.persistence.PersistenceEngine;
import gov.bnl.gums.userGroup.ManualUserGroup;
import gov.bnl.gums.util.QuietLog;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Properties;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.digester.Digester;
import org.apache.log4j.Logger;


/** 
 * Facade for the whole business logic available in GUMS. Using GUMS means
 * instanciating an object of this class, and use it to reach the rest of the
 * functionalities.
 *
 * @author  Gabriele Carcassi, Jay Packard
 */
public class GUMS {
    static final public String siteAdminLogName = "gums.siteAdmin";
    static final public String gumsAdminLogName = "gums.gumsAdmin";
//    static private Logger log = Logger.getLogger(GUMS.class); // only use this log for particularly tricky aspects of this class - otherwise log within lower level classes
    static public QuietLog gumsAdminEmailLog = new QuietLog("gums.gumsAdminEmail");
    static protected Logger gumsAdminLog = Logger.getLogger(GUMS.gumsAdminLogName);
    static protected String version = null;
    
    static public String getMajorMinorVersion() {
    	String mmVersion = getVersion();
    	if (mmVersion != null) {
	    	int dotIndex = mmVersion.lastIndexOf('.');
	    	if (dotIndex != -1) {
	    		return mmVersion.substring(0, dotIndex);
	    	}
    	}
    	return "";
    }
    
    static public String getVersion() {
    	if (version==null) {
			String pomFile;
    		if (CertCache.getMetaDir() != null) {
    			pomFile = CertCache.getMetaDir()+"/maven/gov.bnl.racf/gums-service/pom.xml";
    		}
    		else {
    			// For testing
    			String curDir = System.getProperty("user.dir");
    			if (curDir.endsWith("/WEB-INF/scripts"))
    				curDir += "/../../../..";
    			pomFile = curDir + "/pom.xml";
    			
    			// for command line scripts
    			if (!new File(pomFile).exists()) {
        			curDir = System.getProperty("user.dir");
        			if (curDir.endsWith("/WEB-INF/scripts"))
        				curDir += "/../../META-INF/maven/gov.bnl.racf/gums-service";
        			pomFile = curDir + "/pom.xml";    				
    			}
    		}
    		
			Digester digester = new Digester();
			digester.addObjectCreate("project/version", Version.class);
			digester.addCallMethod("project/version","setVersion",0);
			Version versionCls = null;
			try {
				versionCls = (Version)digester.parse("file://"+pomFile);
			} catch (Exception e) {
				gumsAdminLog.warn("Cannot get GUMS version from "+pomFile);
				return "?";
			}
			if (versionCls == null) {
				gumsAdminLog.warn("Cannot get GUMS version from "+pomFile);
				return "?";
			}
			else
				gumsAdminLog.debug("Loaded GUMS version " + versionCls.getVersion() + " from '" + pomFile + "'" );
			version = versionCls.getVersion();
		}
		return version;
    }

    static TimerTask updateTask = new TimerTask() {
        public void run() {
        	synchronized(this) {
                try {
                    gumsAdminLog.info("Starting automatic updateGroups");
                    gums.beginTransaction();
                    gums.getConfiguration().updateUserGroupMembers();
                    gums.commit();
                    gumsAdminLog.info("Automatic updateGroups ended");
                } catch (Exception e) {
                	gums.rollback();
                    gumsAdminLog.warn("Automatic group update had failures - " + e.getMessage());
                    gumsAdminEmailLog.put("updateUserGroup", e.getMessage(), false);
                }
        	}
        }
    };     
    
    static TimerTask updateBannedTask = new TimerTask() {
        public void run() {
        	synchronized(gums) {
                try {
                    gumsAdminLog.info("Starting automatic updateBannedGroups");
                    gums.beginTransaction();
                    gums.getConfiguration().updateBannedUserGroupMembers();
                    gums.commit();
                    gumsAdminLog.info("Automatic updateBannedGroups ended");
                } catch (Exception e) {
                	gums.rollback();
                    gumsAdminLog.warn("Automatic banned group update had failures - " + e.getMessage());
                    gumsAdminEmailLog.put("updateBannedGroups", e.getMessage(), false);
                }
        	}
        }
    };
    
    static TimerTask emailWarningTask = new TimerTask() {
        public void run() {
            synchronized(gums) {
          	   if (gumsAdminEmailLog.hasMessages()) {
             	   gumsAdminEmailLog.logMessages();
          	   }
            }
         }
     };  
    
    static TimerTask autoMergeTask = new TimerTask() {
        public void run() {
           synchronized(gums) {
        	   try {
                   gumsAdminLog.info("Starting automatic merging");
                   gums.beginTransaction();
                   Configuration conf = (Configuration)gums.getConfiguration().clone();
                   if (conf!=null && conf.isAutoMerge() && 
                		   conf.getMergeURIs()!=null && !conf.getMergeURIs().isEmpty() &&
                		   conf.getMergeHostToGroupMapping()!=null && !conf.getMergeHostToGroupMapping().isEmpty()) {
                	   Configuration newConfig = ConfigurationToolkit.parseConfigurationTemplates(conf.getMergeURIs(), null);
                	   conf.mergeConfigurationTemplate(newConfig, conf.getMergeHostToGroupMapping(), true);
                	   gums.setConfiguration(conf);
                   }
                   gums.commit();
                   gumsAdminLog.info("Automatic merging ended");
               } 
        	   catch (Exception e) {
                   gums.rollback();
                   gumsAdminLog.warn("Automatic merge had failures - " + e.getMessage());
                   gumsAdminEmailLog.put("updateUserGroup", e.getMessage(), false);
               }
           }
        }
    };
    
    /**
     * Create a thread that updates user group membership every so often
     * 
     * @param gums
     */
    static private synchronized void startWorkerThread(final GUMS gums) {
    	Properties properties = gums.getProperties();
    	
        // If JNDI property is set, run the update every x minutes
        if (gums.timer == null) {
        	// update groups
       	
        	gums.timer = new Timer();
        	
        	// update user groups
        	Integer updateMinutes = Integer.parseInt(properties.getProperty("updateGroupsMinutes", "720"));
            gums.timer.scheduleAtFixedRate(updateTask, 10000, updateMinutes.intValue()*60*1000);
            gumsAdminLog.info("Automatic group update set: will refresh every " + updateMinutes.intValue() + " minutes starting in 10 seconds.");
            
            // update banned groups
        	Integer updateBannedMinutes = Integer.parseInt(properties.getProperty("updateBannedGroupsMinutes", "60"));     
            gums.timer.scheduleAtFixedRate(updateBannedTask, 10000, updateBannedMinutes.intValue()*60*1000);
            gumsAdminLog.info("Automatic banned group update set: will refresh every " + updateBannedMinutes.intValue() + " minutes starting in 10 seconds.");
            
            // sparse email logger
        	Integer emailWarningHours = Integer.parseInt(properties.getProperty("emailWarningHours", "48"));     	
            gums.timer.scheduleAtFixedRate(emailWarningTask, 5*60*1000, emailWarningHours.intValue()*60*60*1000);
            gumsAdminLog.info("Automatic email warning set: will refresh every " + emailWarningHours.intValue() + " hours starting in 5 minutes.");
            
	        // auto merge
        	Integer autoMergeHours = Integer.parseInt(properties.getProperty("autoMergeHours", "24"));    	
            gums.timer.scheduleAtFixedRate(autoMergeTask, 0, autoMergeHours.intValue()*60*60*1000);
            gumsAdminLog.info("Automatic merging set: will refresh every " + autoMergeHours.intValue() + " hours starting now.");
            
            // http proxy host
			String httpProxyHost = properties.getProperty("http.proxyHost");
			if (httpProxyHost!=null && httpProxyHost.length()>0)
				System.setProperty("http.proxyHost", httpProxyHost);
            
            // http proxy port
			String httpProxyPort = properties.getProperty("http.proxyPort");
			if (httpProxyPort!=null && httpProxyPort.length()>0)
				System.setProperty("http.proxyPort", httpProxyPort);
            
            // https proxy host
			String httpsProxyHost = properties.getProperty("https.proxyHost");
			if (httpsProxyHost!=null && httpsProxyHost.length()>0)
				System.setProperty("https.proxyHost", httpsProxyHost);
            
            // https proxy port
			String httpsProxyPort = properties.getProperty("https.proxyPort");
			if (httpsProxyPort!=null && httpsProxyPort.length()>0)
				System.setProperty("https.proxyPort", httpsProxyPort);
        }
    }
    
    protected Timer timer;
    protected PersistenceEngine persEngine;
    protected CoreLogic coreLogic = new CoreLogic();
    protected Properties properties = null;
    protected static GUMS gums = null;
    
    public GUMS() throws FileNotFoundException, IOException, IllegalArgumentException, SecurityException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException {
    	properties = new Properties();
    	properties.load(new FileInputStream(CertCache.getConfigDir() + "/gums.properties"));
    	
    	if (properties.get("persistenceEngineClass") != null) {
    		persEngine = (PersistenceEngine)Class.forName(properties.get("persistenceEngineClass").toString()).getConstructor().newInstance();
    	}
    	else
    		persEngine = new HibernatePersistenceEngine(properties);
        
        startWorkerThread(this);
    }
    
    public GUMS(Properties properties) throws IllegalArgumentException, SecurityException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException {
    	this.properties = properties;
    	
    	if (properties!=null && properties.get("gums.persistenceEngineClass") != null)
			persEngine = (PersistenceEngine)Class.forName(properties.get("gums.persistenceEngineClass").toString()).getConstructor(Properties.class).newInstance(properties);
    	else
    		persEngine = new HibernatePersistenceEngine(properties);
        
        startWorkerThread(this);
    }
    
    public Configuration getConfiguration() {
    	Configuration conf = persEngine.retrieveCurrentConfiguration();
		
		return conf;
    }
    
    /**
     * Retrieve the Core Logic to perform actions on the business logic.
     * 
     * @return the resource manager.
     */
    public CoreLogic getCoreLogic() {
        return coreLogic;
    }
    
    static public GUMS getSingletonGums() {
    	try {
			if (gums == null)
				gums = new GUMS();
		} catch (Exception e) {
			gumsAdminLog.fatal(e.getMessage());
			throw new RuntimeException(e);
		}
    	return gums;
    }
        
    public Properties getProperties() {
		return properties;
	}
    
    public void removeBackupConfiguration(String name) {
    	persEngine.removeBackupConfiguration(name);
    }
    
    public Configuration restoreConfiguration(String name) throws Exception {
    	Configuration conf = (Configuration)persEngine.restoreConfiguration(name);
    	gums.timer.schedule(updateTask, 5000);
    	return conf;
    }

	public void setConfiguration(Configuration conf) {
		if (conf.isBackup())
			conf.clearCachedItems();
		persEngine.setConfiguration(conf);
    }

	public Collection<String> getBackupConfigurationNames() {
		return persEngine.retrieveBackupConfigurationNames();
	}
	
	public void commit() {
		persEngine.commit();
	}
	
	public void beginTransaction() {
		persEngine.beginTransaction();
	}
   
	public void rollback() {
		persEngine.rollback();
	}
	
}
