/*
 * ManualAccountMapper.java
 *
 * Created on May 25, 2004, 5:06 PM
 */

package gov.bnl.gums.account;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.OneToMany;
import javax.persistence.Transient;

import gov.bnl.gums.Mapping;
import gov.bnl.gums.SiteUser;
import gov.bnl.gums.configuration.ConfigElement;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.util.RWLock;

/** 
 * An account mapping policy that looks at a stored table to determine the
 * account. The database implementation is an abstract interface, which
 * allows the mapping to be persistent on different mediums (i.e. database,
 * LDAP, file, ...)
 * <p>
 * This class will provide also configurable data caching.
 *
 * @author Gabriele Carcassi, Jay Packard
 */
@Entity
public class ManualAccountMapper extends AccountMapper {
    private Map<String, SiteUser> map = new HashMap<String, SiteUser>();
    private Map<SiteUser, Set<String>> reverseMap = new HashMap<SiteUser, Set<String>>();
	protected RWLock patternRWLock = new RWLock();

	// persistent variables
	protected Date lastUpdate;
	protected List<Mapping> mappings = new ArrayList<Mapping>();
	
    public ManualAccountMapper() {
    	super();
    }
    
    public ManualAccountMapper(Configuration configuration, String name) {
    	super(configuration, name);
    }
	
    public ConfigElement clone(Configuration conf, boolean allowReplace) {
    	ManualAccountMapper manualAccountMapper = (ManualAccountMapper)super.clone(conf, allowReplace);
    	manualAccountMapper.setMappings(new ArrayList<Mapping>(mappings));
    	return manualAccountMapper;
    }
    
	public final void createMapping(String dn, SiteUser siteUser) {
		synchronized(mappings) {
			Mapping mapping = null;
			int index;
			if ((index = mappings.indexOf(siteUser)) != -1) {
				mapping = mappings.get(index);
				mapping.setDn(dn);
			}
			else {
				mapping = new Mapping(dn, siteUser);
				mappings.add(mapping);
			}
			mapping.setAccountMapper(this);
		}
		refreshMaps();
	}
	
	public Date getLastUpdate() {
		return lastUpdate;
	}
	
	@OneToMany(cascade=CascadeType.ALL, mappedBy="accountMapper") 
	@org.hibernate.annotations.Cascade(org.hibernate.annotations.CascadeType.DELETE_ORPHAN)
	public List<Mapping> getMappings() {
		return mappings;
	}
    
    public SiteUser mapDn(String dn, boolean createNew){
		if (dn == null)
			return null;
    	try {
	    	patternRWLock.getReadLock();
			return map.get(dn);
    	}
    	finally {
    		patternRWLock.releaseLock();
    	}
	}
    
    public Set<String> mapSiteUser(SiteUser siteUser) {
    	if (siteUser == null)
			return null;
    	try {
	    	patternRWLock.getReadLock();
			return reverseMap.get(siteUser);
    	}
    	finally {
    		patternRWLock.releaseLock();
    	}
    }

    public boolean removeMapping(String dn) {
    	boolean retVal = false;
    	assert(dn != null);
    	synchronized(mappings) {
			for (Mapping m: mappings) {
				if (m.getDn().equals(dn)) {
					retVal = mappings.remove(m);
					break;
				}
			}
    	}
    	if (retVal)
    		refreshMaps();
    	return retVal;
	}
    
    public void removeMapping(SiteUser siteUser) {
		assert(siteUser != null);
		synchronized(mappings) {
			for (Mapping m: mappings) {
				if (m.getSiteUser().equals(siteUser)) {
					mappings.remove(m);
				}
			}
		}
		refreshMaps();
	}      

	public void setLastUpdate(Date lastUpdate) {
		this.lastUpdate = lastUpdate;
	}
	
	
	protected void refreshMaps() {
		try {
	    	patternRWLock.getWriteLock();
	    	map.clear();
	    	reverseMap.clear();
			synchronized (mappings) {
				for (Mapping m: mappings) {
					if (m.getDn() != null) {
						map.put(m.getDn(), m.getSiteUser());
						Set<String> users = reverseMap.get(m.getSiteUser());
						if (users == null) {
							users = new HashSet<String>();
							users.add(m.getDn());
							reverseMap.put(m.getSiteUser(), users);
						}
					}
				}
			}
    	}
    	finally {
    		patternRWLock.releaseLock();
    	}
	}

	private void setMappings(List<Mapping> mappings) {
		if (mappings == null || mappings.size()==0)
			return;
		this.mappings = mappings;
    	for (Mapping m: mappings)
    		m.setAccountMapper(this);
		refreshMaps();
	}
   
}
