package gov.bnl.gums.configuration;

import gov.bnl.gums.util.ClassUtils;
import gov.bnl.gums.util.StringUtil;

import java.lang.annotation.*;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.lang.reflect.Method;

import javax.annotation.processing.*;

import javax.persistence.*;

@MappedSuperclass
@SupportedAnnotationTypes({"gov.bnl.gums.configuration.ConfigElement.PermanentField"})
public abstract class ConfigElement {
	@Retention(RetentionPolicy.RUNTIME)
	@Target({ElementType.METHOD})
	public @interface ConfigFieldAnnotation {
		String label();
		String example() default "";
		String help() default "";
		int size() default 32;
		String choices() default "";
		int order() default -1;
		boolean orderedList() default false;
		boolean showInConciseView() default false;
	}
	
	public class ConfigField {
		public String setterName;
		public String getterName;
		public String configFieldName;
		public Class<?> type;
		public ConfigFieldAnnotation configFieldAnnotation;
		public Object value;
	}
	
	List<ConfigField> configurationFields;
	protected WeakReference<Configuration> configurationRef = null;

	// persistent variables
	protected String name;
	protected String description;
	protected String source;
	private int id;

	public ConfigElement() {
	}

	public ConfigElement(Configuration configuration, String name) {
		configurationRef = new WeakReference<Configuration>(configuration);
		setName(name);
		configuration.addConfigElement(this);
	}

	public ConfigElement clone(Configuration configuration, boolean allowReplace) {
		ConfigElement configElement = null;
		try {
			configElement = this.getClass().getConstructor().newInstance();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		configElement.setConfiguration(configuration);
		configElement.setName(name);
		configuration.addConfigElement(this, allowReplace);

		for (ConfigField field : getConfigFields()) {
			try {
				getClass().getMethod(field.setterName, String.class).invoke(configElement, field.value);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
		return configElement;
	}

	public boolean equals(Object obj) {
		if (obj instanceof ConfigElement) {
			ConfigElement ce = (ConfigElement) obj;
			Iterator<ConfigField> it = getConfigFields().iterator();
			while (it.hasNext()) {
				ConfigField field = it.next();
				Object value1 = null;
				Object value2 = null;
				try {
					value1 = getClass().getMethod(field.getterName).invoke(this);
					value2 = getClass().getMethod(field.getterName).invoke(ce);
				} catch (Exception e) {
					assert(false);
				}
				if ((value1 == null && value2 != null)
						|| (value1 != null && value2 == null)) {
					return false;
				}
				if (value1 != null && value2 != null && !value1.equals(value2)) {
					return false;
				}
			}
			return true;
		}
		return false;
	}

	@ManyToOne
	@JoinColumn(name = "configuration")
	public Configuration getConfiguration() {
		return configurationRef != null ? configurationRef.get() : null;
	}
	
	@Transient
	public List<ConfigField> getConfigFields() {
		if ((getConfiguration()!=null && !getConfiguration().isReadOnly()) || configurationFields==null) {
			List<ConfigField> tempConfigurationFields = new ArrayList<ConfigField>();
			Method[] methods = getClass().getMethods();
			for (Method method : methods) {
				ConfigField field = new ConfigField();
				ConfigFieldAnnotation configFieldAnnotation = method.getAnnotation(ConfigFieldAnnotation.class);
				if (configFieldAnnotation != null) {
					field.configFieldAnnotation = configFieldAnnotation;
					field.type = method.getParameterTypes()[0];
					field.setterName = method.getName();
					field.getterName = field.setterName.replace("set", "get");
					try {
						try {
							field.value = getClass().getMethod(field.getterName, (Class<?>[])null).invoke(this, (Object[])null);
						} catch (NoSuchMethodException e) {
							field.getterName = field.setterName.replace("set", "is");
							field.value = getClass().getMethod(field.getterName, (Class<?>[])null).invoke(this, (Object[])null);
						}
						String firstChar = field.setterName.substring(3, 4);
						field.configFieldName = field.setterName.replace("set"+firstChar, firstChar.toLowerCase());
						tempConfigurationFields.add(field);
					}
					catch (Exception e) {
						throw new RuntimeException(e.getMessage());
					}
				}
			}
			// order
			configurationFields = new ArrayList<ConfigField>();
			int max;
			do {
				max = -1;
				int index = -1;
				for (int i=0; i<tempConfigurationFields.size(); i++) {
					ConfigField configField = tempConfigurationFields.get(i);
					int order = configField.configFieldAnnotation.order();
					if (order > max) {
						max = order;
						index = i;
					}
				}
				if (index!=-1) {
					configurationFields.add(0, tempConfigurationFields.get(index));
					tempConfigurationFields.remove(index);
				}
			} while (max > -1);
			for (int i=0; i<tempConfigurationFields.size(); i++)
				configurationFields.add(tempConfigurationFields.get(i));
		}
		return configurationFields;
	}

	public String getDescription() {
		return description;
	}

	public String getName() {
		return name;
	}

	public String getSource() {
		return source;
	}

	public int hashCode() {
		return getName()==null ? 0 : getName().hashCode();
	}

	public void setConfiguration(Configuration configuration) {
		configurationRef = new WeakReference<Configuration>(configuration);
	}

	@ConfigFieldAnnotation(label="Description", showInConciseView=true, order=2)
	public void setDescription(String description) {
		this.description = description;
	}

	@ConfigFieldAnnotation(label="Name", showInConciseView=true, order=1)
	public void setName(String name) {
		this.name = name;
	}

	@ConfigFieldAnnotation(label="Source", example="OSG", showInConciseView=true, order=3)
	public void setSource(String source) {
		this.source = source;
	}

	@SuppressWarnings("unchecked")
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(ClassUtils.getClassName(this.getClass()) + ": ");
		Iterator<ConfigField> it = getConfigFields().iterator();
		while (it.hasNext()) {
			ConfigField field = it.next();
			if (field.value instanceof String)
				sb.append(field.configFieldName + "='" + field.value + "'");
			else if (field.configFieldName.equals("properties")) {
				Properties properties = (Properties)field.value;
				for (Object property : properties.keySet())
					sb.append(property + "='" + properties.get(property) + "'");
			} else if (field.value instanceof Collection) {
				sb.append(field.configFieldName + "='");
				Iterator<String> it2 = ((Collection<String>) field.value).iterator();
				while (it2.hasNext()) {
					sb.append(it2.next());
					if (it2.hasNext())
						sb.append(", ");
				}
				sb.append("'");
			}
			if (it.hasNext())
				sb.append(", ");
		}
		return sb.toString();
	}

	/**
	 * Get XML output of this element
	 * 
	 * @param className
	 * @return String
	 */
	@SuppressWarnings("unchecked")
	final public String toXML() {
		List<ConfigField> fields = getConfigFields();
		StringBuffer sb = new StringBuffer();
		sb.append("\t\t<"
				+ classNameToXmlElement(ClassUtils.getClassName(this.getClass()))
				+ (fields.size() > 0 ? "\n" : ""));
		Iterator<ConfigField> it = fields.iterator();
		while (it.hasNext()) {
			ConfigField field = it.next();
			if (field.configFieldName.equals("properties")) {
				Properties properties = (Properties) field.value;
				for (Object property : properties.keySet()) {
					Object value = properties.get(property);
					sb.append("\t\t\t" + property + "='" + (value!=null?value:"") + "'");
				}
			} 
			else if (field.value instanceof Collection) {
				sb.append("\t\t\t"+ field.configFieldName+ "='"+ StringUtil.toCommaSeparatedList((Collection<String>) field.value) + "'");
			}
			else {
				sb.append("\t\t\t" + field.configFieldName + "='" + (field.value!=null?field.value.toString():"") + "'");
			}
			if (it.hasNext())
				sb.append("\n");
		}
		sb.append("/>\n\n");
		return sb.toString();
	}
	
	protected String classNameToXmlElement(String className) {
		String retVal = className.substring(0, 1).toLowerCase() + className.substring(1);
		retVal = retVal.replace("lDAP", "ldap");
		retVal = retVal.replace("vOMS", "voms");
		return retVal;
	}
	
	@Id
	@GeneratedValue(strategy = GenerationType.AUTO)
	@SuppressWarnings("unused")
	private int getId() {
		return id;
	}

	@SuppressWarnings("unused")
	private void setId(int id) {
		this.id = id;
	}
}
