/*
 * Configuration.java
 *
 * Created on May 24, 2004, 2:33 PM
 */

package gov.bnl.gums.configuration;

import gov.bnl.gums.account.AccountMapper;
import gov.bnl.gums.groupToAccount.GroupToAccountMapping;
import gov.bnl.gums.hostToGroup.HostToGroupMapping;
import gov.bnl.gums.userGroup.UserGroup;
import gov.bnl.gums.util.StringUtil;
import gov.bnl.gums.GUMS;
import gov.bnl.gums.GridUser;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.*;

import org.apache.log4j.Logger;

import gov.bnl.gums.persistence.PersistenceEngine;

import javax.persistence.*;

/** 
 * Holds the configuration of GUMS, including which policies will be used
 * for which hosts, which database layer is going to be used and so on.
 * <p>
 * The configuration object will be constructed programmatically by reading
 * an xml file.
 * <p>
 * This class is not thread safe so be sure to not modify during runtime
 *
 * @author Gabriele Carcassi, Jay Packard
 */
@Entity
public class Configuration {
	protected class UpdateThread implements Runnable {
		protected UserGroup userGroup;
		protected boolean onlyUpdatedBannedUsers;
		
		protected UpdateThread(UserGroup userGroup, boolean onlyUpdatedBannedUsers) {
			this.userGroup = userGroup;
			this.onlyUpdatedBannedUsers = onlyUpdatedBannedUsers;
		}
		
	    public void run() {
	    	updaterInstanceCount++;
	    	try {
				userGroup.updateMembers();
				userGroup.setLastUpdate(new Date());
                adminLog.info("User group update for " + userGroup.getName() + " (" + userGroup.getMembers().size() + " users).");
			} catch (Exception e) {
				adminLog.warn("User group update for " + userGroup.getName() + " failed: " + e.getMessage());
		    	failedGroups.append( userGroup.getName() + " - " + e.getMessage() + "\n" );
			} finally {
		    	updaterInstanceCount--;
			}
	    }
	}
	
	static final public int maxUpdaterThreads = 10;
	static private Logger adminLog = Logger.getLogger(GUMS.gumsAdminLogName);
	static private Logger log = Logger.getLogger(Configuration.class);
	static public int instanceCount = 0;
	static public int updaterInstanceCount = 0;
	
	static protected String toCommaSepList(Collection<String> collection) {
		StringBuffer sb = new StringBuffer();
		Iterator<String> it = collection.iterator();
		while (it.hasNext()) {
			String item = it.next();
			sb.append(item);
			if (it.hasNext())
				sb.append(", ");
		}
		return sb.toString();
	}
	protected boolean versionWasSet = false;
	protected boolean isMetaOnly = false;
	protected String source = null; // only used for template configurations - not persisted
	protected boolean transformingFromOld11Version = false;
	protected List<String> bannedUserGroupsList = new ArrayList<String>();
	
	protected boolean readOnly = false;
	// persistent variables
	private int id;
	protected String name = null;
	protected String version = GUMS.getVersion(); // default
	protected boolean isTemplate = false; // flag indicating config without persistence factories or hostToGroup mappings
	protected boolean allowGridmapFiles = true;
	protected boolean autoMerge = false;
	protected String mergeHostToGroupMapping;
	protected String mergeURIs;
	protected String sslCAFiles;
	protected String sslCertfile = "/etc/grid-security/http/httpcert.pem";
	protected String sslKey = "/etc/grid-security/http/httpkey.pem";
	protected String sslKeyPasswd;
	protected String bannedUserGroups;
	protected Map<String, AccountMapper> accountMappers = new TreeMap<String, AccountMapper>();
	protected Map<String, UserGroup> userGroups = new TreeMap<String, UserGroup>();
	protected Map<String, GroupToAccountMapping> groupToAccountMappings = new TreeMap<String, GroupToAccountMapping>();
	protected List<HostToGroupMapping> hostToGroupMappings = new ArrayList<HostToGroupMapping>();
	protected StringBuffer failedGroups = new StringBuffer();

	protected Configuration thisConfig = null;;

	public Configuration() {
		instanceCount++;
		this.thisConfig = this;
	}

	public Configuration(boolean transformingFromOld11Version) {
		this.transformingFromOld11Version = transformingFromOld11Version;
		this.thisConfig = this;
		instanceCount++;
	}
	
	public void addAccountMapper(AccountMapper accountMapper) {
		addAccountMapper(accountMapper, false);
	}
	
	public void addAccountMapper(AccountMapper accountMapper, boolean allowReplace) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");
		
		log.trace("Adding account mapper to the configuration: " + accountMapper.getName());
		if (accountMappers.get(accountMapper.getName())!=null && !allowReplace) {
			if (transformingFromOld11Version)
				log.trace("Replacing account mapper: " + accountMapper.getName());
			else {
				String message = "Account mapper " + accountMapper.getName() + " already exists";
				log.debug(message);
				adminLog.error(message);
				throw new RuntimeException(message);
			}
		}     	
		accountMappers.put(accountMapper.getName(), accountMapper);
		if (accountMapper.getConfiguration()==null)
			accountMapper.setConfiguration(this);
		if (isTemplate && source != null)
			accountMapper.setSource(source);
	}
	
	public void addConfigElement(ConfigElement configElement, boolean allowReplace) {
		if (configElement instanceof AccountMapper)
			addAccountMapper((AccountMapper)configElement, allowReplace);
		else if (configElement instanceof UserGroup)
			addUserGroup((UserGroup)configElement, allowReplace);
		else if (configElement instanceof GroupToAccountMapping)
			addGroupToAccountMapping((GroupToAccountMapping)configElement, allowReplace);
		else if (configElement instanceof HostToGroupMapping)
			addHostToGroupMapping((HostToGroupMapping)configElement);
	}
	
	public void addConfigElement(ConfigElement configElement) {
		addConfigElement(configElement, false);
	}
	
	public void addGroupToAccountMapping(GroupToAccountMapping g2AMapping) {
		addGroupToAccountMapping(g2AMapping, false);
	}
		
	public void addGroupToAccountMapping(GroupToAccountMapping g2AMapping, boolean allowReplace) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		log.trace("Adding group to account mapping to the configuration: " + g2AMapping.getName());
		if (getGroupToAccountMapping(g2AMapping.getName())!=null && !allowReplace) {
			String message = "Group to account mapping " + g2AMapping.getName() + " already exists";
			log.debug(message);
			adminLog.error(message);
			throw new RuntimeException(message);
		}
		groupToAccountMappings.put(g2AMapping.getName(), g2AMapping);
		if (g2AMapping.getConfiguration()==null)
			g2AMapping.setConfiguration(this);
		if (isTemplate && source != null)
			g2AMapping.setSource(source);
	}
	
	public void addHostToGroupMapping(HostToGroupMapping h2GMapping) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		log.trace("Adding Host to group mapping to the configuration: " + h2GMapping.getName());
		if (getHostToGroupMapping(h2GMapping.getName()) != null) {
			String message = "Host to group mapping " + h2GMapping.getName() + " already exists";
			log.debug(message);
			adminLog.error(message);
			throw new RuntimeException(message);
		}
		hostToGroupMappings.add(h2GMapping);
		if (h2GMapping.getConfiguration()==null)
			h2GMapping.setConfiguration(this);
	}

	public void addHostToGroupMapping(int index, HostToGroupMapping h2GMapping) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		log.trace("Adding Host to group mapping to the configuration: " + h2GMapping.getName());
		if (getHostToGroupMapping(h2GMapping.getName()) != null) {
			String message = "Host to group mapping " + h2GMapping.getName() + " already exists";
			log.debug(message);
			adminLog.error(message);
			throw new RuntimeException(message);
		}
		hostToGroupMappings.add(index, h2GMapping);
		if (h2GMapping.getConfiguration()==null)
			h2GMapping.setConfiguration(this);
	}

	public void addPersistenceFactory(PersistenceEngine peristenceFactory) {
	}

	public void addUserGroup(UserGroup userGroup) {
		addUserGroup(userGroup, false);
	}

	public void addUserGroup(UserGroup userGroup, boolean allowReplace) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		log.trace("Adding user group to the configuration: " + userGroup.getName());
		if (userGroups.get(userGroup.getName())!=null && !allowReplace) {
			String message = "User group " + userGroup.getName() + " already exists";
			log.debug(message);
			adminLog.error(message);
			throw new RuntimeException(message);
		}             
		userGroups.put(userGroup.getName(), userGroup);
		if (userGroup.getConfiguration()==null)
			userGroup.setConfiguration(this);
		if (isTemplate && source != null)
			userGroup.setSource(source);
	}    

	public void clearCachedItems() {
		for (UserGroup ug: userGroups.values())
			ug.clearCachedItems();
	}

	// Make sure to not reference any of the old configuration's data
	// or else the old configuration will never be freed
	public Object clone() {
		Configuration newConf = new Configuration();

		newConf.setVersion( getVersion() );
		newConf.setAllowGridmapFiles( isAllowGridmapFiles() );
		newConf.setBannedUserGroups( getBannedUserGroups() );
		newConf.setSslCAFiles(getSslCAFiles());
		newConf.setSslCertfile(getSslCertfile());
		newConf.setSslKey(getSslKey());
		newConf.setSslKeyPasswd(getSslKeyPasswd());
		newConf.setIsTemplate(isTemplate());
		newConf.setSource(getSource());
		newConf.setAutoMerge(isAutoMerge());
		newConf.setMergeHostToGroupMapping(getMergeHostToGroupMapping());
		newConf.setMergeURIs(getMergeURIs());

		Iterator<AccountMapper> amIt = accountMappers.values().iterator();
		while (amIt.hasNext() )
			amIt.next().clone(newConf, false);    

		Iterator<UserGroup> ugIt = userGroups.values().iterator();
		while (ugIt.hasNext() )
			ugIt.next().clone(newConf, false);  

		Iterator<GroupToAccountMapping> g2aIt = groupToAccountMappings.values().iterator();
		while (g2aIt.hasNext() )
			g2aIt.next().clone(newConf, false);  

		Iterator<HostToGroupMapping> h2gIt = hostToGroupMappings.iterator();
		while (h2gIt.hasNext() )
			h2gIt.next().clone(newConf, false); 

		return newConf;
	}
	
	public boolean equals(Object obj) {
		if (obj instanceof Configuration) {
			if (((Configuration)obj).getName()==null && getName()==null)
				return true;
			if (((Configuration)obj).getName()==null || getName()==null)
				return false;
			else
				return ((Configuration)obj).getName().equals(getName());
		}
		else
			throw new RuntimeException("cannot compare configuration to :" + obj.getClass().toString());
	}

	public AccountMapper getAccountMapper(String accountMapper)  {
		return (AccountMapper)accountMappers.get(accountMapper);
	}

	@OneToMany(cascade=CascadeType.ALL, fetch=FetchType.EAGER, mappedBy="configuration")
	@org.hibernate.annotations.Cascade(org.hibernate.annotations.CascadeType.DELETE_ORPHAN)
	public Map<String, AccountMapper> getAccountMappers()  {
		return accountMappers;
	}

	public String getBannedUserGroups() {
		return bannedUserGroups;
	}

	public GroupToAccountMapping getGroupToAccountMapping(String groupToAccountMapping) {
		return (GroupToAccountMapping)groupToAccountMappings.get(groupToAccountMapping);
	}

	@OneToMany(cascade=CascadeType.ALL, fetch=FetchType.EAGER, mappedBy="configuration") 
	@org.hibernate.annotations.Cascade(org.hibernate.annotations.CascadeType.DELETE_ORPHAN)
	public Map<String, GroupToAccountMapping> getGroupToAccountMappings() {
		return groupToAccountMappings;
	}

	public HostToGroupMapping getHostToGroupMapping(String name) {
		Iterator<HostToGroupMapping> it = hostToGroupMappings.iterator();
		while(it.hasNext()) {
			HostToGroupMapping hostToGroupMapping = (HostToGroupMapping)it.next();
			if(hostToGroupMapping.getName().equals(name))
				return hostToGroupMapping;
		}
		return null;
	}

	public List<ConfigElement> getConfigElements(Class<?> cls) {
		if (cls.equals(AccountMapper.class))
			return new ArrayList<ConfigElement>(getAccountMappers().values());
		else if (cls.equals(UserGroup.class))
			return new ArrayList<ConfigElement>(getUserGroups().values());
		else if (cls.equals(GroupToAccountMapping.class))
			return new ArrayList<ConfigElement>(getGroupToAccountMappings().values());
		else if (cls.equals(HostToGroupMapping.class))
			return new ArrayList<ConfigElement>(getHostToGroupMappings());
		else
			assert(false);
		return null;
	}
	
	public ConfigElement getConfigElement(Class<?> cls, String name) {
		if (cls.equals(AccountMapper.class))
			return getAccountMapper(name);
		else if (cls.equals(UserGroup.class))
			return getUserGroup(name);
		else if (cls.equals(GroupToAccountMapping.class))
			return getGroupToAccountMapping(name);
		else if (cls.equals(HostToGroupMapping.class))
			return getHostToGroupMapping(name);
		else
			assert(false);
		return null;
	}
	
	@OneToMany(cascade=CascadeType.ALL, fetch=FetchType.EAGER, mappedBy="configuration") 
	@org.hibernate.annotations.Cascade(org.hibernate.annotations.CascadeType.DELETE_ORPHAN)
	public List<HostToGroupMapping> getHostToGroupMappings() {
		return hostToGroupMappings;
	}

	public String getMergeHostToGroupMapping() {
		return mergeHostToGroupMapping;
	}    

	public String getMergeURIs() {
		return mergeURIs;
	}    

	public String getName() {
		return name;
	}

	/**
	 * Returns a list of all the readers with all read access defined in the configuration file.
	 *
	 * @return a list of ReaderGroupManager objects.
	 */
	@Transient
	public ArrayList<UserGroup> getReadAllUserGroups() {
		ArrayList<UserGroup> readers = new ArrayList<UserGroup>();
		Iterator<UserGroup> it = userGroups.values().iterator();
		while( it.hasNext() ) {
			UserGroup userGroup = it.next();
			if( userGroup.hasReadAllAccess() ) {
				readers.add(userGroup);
			}
		}
		return readers;
	}

	/**
	 * Returns a list of all the self readers defined in the configuration file.
	 * 
	 * @return a list of ReaderGroupManager objects.
	 */
	@Transient
	public ArrayList<UserGroup> getReadSelfUserGroups() {
		ArrayList<UserGroup> readers = new ArrayList<UserGroup>();
		Iterator<UserGroup> it = userGroups.values().iterator();
		while( it.hasNext() ) {
			UserGroup userGroup = (UserGroup)it.next();
			if( userGroup.hasReadSelfAccess() ) {
				readers.add(userGroup);
			}
		}
		return readers;
	}

	@Transient
	public String getSource() {
		return source;
	}

	public String getSslCAFiles() {
		return sslCAFiles;
	}

	public String getSslCertfile() {
		return sslCertfile;
	}

	public String getSslKey() {
		return sslKey;
	}   

	public String getSslKeyPasswd() {
		return sslKeyPasswd;
	}   

	public UserGroup getUserGroup(String userGroup) {
		return (UserGroup)userGroups.get(userGroup);
	}   

	@OneToMany(cascade=CascadeType.ALL, fetch=FetchType.EAGER, mappedBy="configuration") 
	@org.hibernate.annotations.Cascade(org.hibernate.annotations.CascadeType.DELETE_ORPHAN)
	public Map<String, UserGroup> getUserGroups() {
		return userGroups;
	}

	public String getVersion() {
		return version;
	}

	/**
	 * Returns a list of all the writer groups defined in the configuration file.
	 *
	 * @return a list of WriterGroupManager objects.
	 */
	@Transient
	public ArrayList<UserGroup> getWriteUserGroups() {
		ArrayList<UserGroup> writers = new ArrayList<UserGroup>();
		Iterator<UserGroup> it = userGroups.values().iterator();
		while( it.hasNext() ) {
			UserGroup userGroup = it.next();
			if( userGroup.hasWriteAccess() ) {
				writers.add(userGroup);
			}
		}
		return writers;
	}

	public int hashCode() {
		return getName()==null ? 0 : getName().hashCode();
	}

	public boolean isAllowGridmapFiles() {
		return allowGridmapFiles;
	}

	public boolean isAutoMerge() {
		return autoMerge;
	}

	@Transient 
	public Boolean isBackup() {
		return name != null;
	}

	@Transient 
	public boolean isCurrent() {
		return name == null;
	}

	@Transient
	public boolean isMetaOnly() {
		return isMetaOnly;
	}

	@Transient
	public boolean isReadOnly() {
		return readOnly;
	}

	@Transient
	public boolean isTemplate() {
		return isTemplate;
	}

	public boolean isUserBanned(GridUser user) {
		Iterator<String> it = bannedUserGroupsList.iterator();
		while (it.hasNext()) {
			UserGroup userGroup = getUserGroup(it.next());
			return userGroup.isMember(user);
		}
		return false;
	}

	public void mergeConfigurationTemplate(Configuration configuration, String hostToGroupMappingName, boolean removeDepricated) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		if (!configuration.isTemplate())
			throw new RuntimeException("Can only merge in configuration templates");

		if (configuration.getSource() == null || configuration.getSource().length()==0)
			throw new RuntimeException("Source must be set in configuration templates");
		
		// Get "to" host to group object
		HostToGroupMapping toH2GM = null;
		toH2GM = getHostToGroupMapping(hostToGroupMappingName);
		assert(toH2GM != null);

		// Get "from" host to group object
		assert(configuration.getHostToGroupMappings().size()==1); // should be enforced by schema file
		HostToGroupMapping fromH2GM = (HostToGroupMapping)configuration.getHostToGroupMappings().get(0);
		assert(fromH2GM != null);

		// Remove groupToAccount mappings from toH2GM that will be added back in
		Iterator<String> fromG2AIt = StringUtil.toList(fromH2GM.getGroupToAccountMappings()).iterator();
		while( fromG2AIt.hasNext() ) {
			String fromG2AMName = fromG2AIt.next();
			GroupToAccountMapping toG2AMapping = getGroupToAccountMapping(fromG2AMName);
			if (toG2AMapping != null && toG2AMapping.getSource().equals(configuration.getSource()))
				toH2GM.removeGroupToAccountMapping(fromG2AMName);
		}

		// Remove all elements from elements matching configuration's source
		// if removeDepricated is true - undepricated elements will be added back in
		if (removeDepricated) {
			Iterator<UserGroup> userGroupIt = new ArrayList<UserGroup>(getUserGroups().values()).iterator();
			while( userGroupIt.hasNext() ) {
				UserGroup userGroup = userGroupIt.next();
				if (userGroup.getSource()==null || configuration.getSource()==null)
					throw new RuntimeException("source not set");
				if (userGroup.getSource().equals(configuration.getSource())) {
					removeUserGroup(userGroup.getName(), false);
				}
			}

			Iterator<AccountMapper> accountMapperIt = new ArrayList<AccountMapper>(getAccountMappers().values()).iterator();
			while( accountMapperIt.hasNext() ) {
				AccountMapper accountMapper = accountMapperIt.next();
				if (accountMapper.getSource()==null || configuration.getSource()==null)
					throw new RuntimeException("source not set");
				if (accountMapper.getSource().equals(configuration.getSource()))
					removeAccountMapper(accountMapper.getName(), false);
			}

			Iterator<GroupToAccountMapping> g2AMIt = new ArrayList<GroupToAccountMapping>(getGroupToAccountMappings().values()).iterator();
			while( g2AMIt.hasNext() ) {
				GroupToAccountMapping g2AM = g2AMIt.next();
				if (g2AM.getSource()==null || configuration.getSource()==null)
					throw new RuntimeException("source not set");
				if (g2AM.getSource().equals(configuration.getSource()))
					removeGroupToAccountMapping(g2AM.getName(), false);
			}
		}

		// Loop through groupToAccount mappings
		Iterator<GroupToAccountMapping> fromG2AObjIt = configuration.getGroupToAccountMappings().values().iterator();
		while( fromG2AObjIt.hasNext() ) {
			GroupToAccountMapping fromG2AMapping = fromG2AObjIt.next();
			String fromG2AMName = fromG2AMapping.getName();
			assert(fromG2AMapping != null);
			GroupToAccountMapping toG2AMapping = getGroupToAccountMapping(fromG2AMName);

			boolean shouldInsert = false;
			if (toG2AMapping == null || toG2AMapping.getSource().equals(configuration.getSource()))
				shouldInsert = true;
			
			// Insert or replace user groups
			Iterator<String> userGroupIt = StringUtil.toList(fromG2AMapping.getUserGroups()).iterator();
			while( userGroupIt.hasNext() ) {
				String userGroupName = userGroupIt.next();
				UserGroup fromUserGroup = configuration.getUserGroup(userGroupName);
				assert(fromUserGroup != null);
				UserGroup toUserGroup = getUserGroup(userGroupName);
				if (shouldInsert && (toUserGroup == null || toUserGroup.getSource().equals(configuration.getSource()))) {
					UserGroup clonedUserGroup = (UserGroup)fromUserGroup.clone(this, true);
					clonedUserGroup.setSource(configuration.getSource());
				}
			}

			// Insert or replace account mappers
			Iterator<String> accountMapperIt = StringUtil.toList(fromG2AMapping.getAccountMappers()).iterator();
			while( accountMapperIt.hasNext() ) {
				String accountMapperName = accountMapperIt.next();
				AccountMapper fromAccountMapper = configuration.getAccountMapper(accountMapperName);
				assert(fromAccountMapper != null);
				AccountMapper toAccountMapper = getAccountMapper(accountMapperName);
				if (shouldInsert && (toAccountMapper == null || toAccountMapper.getSource().equals(configuration.getSource()))) {
					AccountMapper clonedAccountMapper = (AccountMapper)fromAccountMapper.clone(this, true);
					clonedAccountMapper.setSource(configuration.getSource());
				}
			}

			// Insert or replace GroupToAccountMapping
			if (shouldInsert) {
				GroupToAccountMapping clonedG2AM = (GroupToAccountMapping)fromG2AMapping.clone(this, true);
				clonedG2AM.setSource(configuration.getSource());

				// Insert GroupToAccountMapping name into HostToGroupMapping
				toH2GM.addGroupToAccountMapping(fromG2AMName);
			}
		}
	}

	public AccountMapper removeAccountMapper(String name, boolean deepRemove) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		if (deepRemove) {
			ArrayList<GroupToAccountMapping> g2amToRemove = new ArrayList<GroupToAccountMapping>();
			for (GroupToAccountMapping g2am: getGroupToAccountMappings().values()) {
				g2am.removeAccountMapper(name);
				if (StringUtil.toList(g2am.getAccountMappers()).size()==0)
					g2amToRemove.add(g2am);
			}
			for (GroupToAccountMapping g2am: g2amToRemove)
				removeGroupToAccountMapping(g2am.getName(), true);
		}
		return (AccountMapper)accountMappers.remove(name);
	}
	
	public ConfigElement removeConfigElement(String name, Class<?> cls, boolean deepRemove) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		if (cls.equals(AccountMapper.class))
			return (ConfigElement)removeAccountMapper(name, deepRemove);
		else if (cls.equals(UserGroup.class))
			return (ConfigElement)removeUserGroup(name, deepRemove);
		else if (cls.equals(GroupToAccountMapping.class))
			return (ConfigElement)removeGroupToAccountMapping(name, deepRemove);
		else if (cls.equals(HostToGroupMapping.class))
			return (ConfigElement)removeHostToGroupMapping(name);
		return null;
	}
	
	public GroupToAccountMapping removeGroupToAccountMapping(String name, boolean deepRemove) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		if (deepRemove) {
			for (HostToGroupMapping h2gm: getHostToGroupMappings())
				h2gm.removeGroupToAccountMapping(h2gm.getName());
		}
		return (GroupToAccountMapping)groupToAccountMappings.remove(name);
	}

	public HostToGroupMapping removeHostToGroupMapping(String name) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		Iterator<HostToGroupMapping> it = hostToGroupMappings.iterator();
		while(it.hasNext()) {
			HostToGroupMapping hostToGroupMapping = it.next();
			if(hostToGroupMapping.getName().equals(name)) {
				hostToGroupMappings.remove( hostToGroupMapping );
				return hostToGroupMapping;
			}
		}
		return null;
	}

	public UserGroup removeUserGroup(String name, boolean deepRemove) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		if (deepRemove) {
			ArrayList<GroupToAccountMapping> g2amToRemove = new ArrayList<GroupToAccountMapping>();
			for (GroupToAccountMapping g2am: getGroupToAccountMappings().values()) {
				g2am.removeUserGroup(name);
				if (StringUtil.toList(g2am.getUserGroups()).size()==0)
					g2amToRemove.add(g2am);
			}
			for (GroupToAccountMapping g2am: g2amToRemove)
				removeGroupToAccountMapping(g2am.getName(), true);
		}
		return (UserGroup)userGroups.remove(name);
	}

	public void setAllowGridmapFiles(boolean allowGridmapFiles) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		this.allowGridmapFiles = allowGridmapFiles;
	}

	public void setAutoMerge(boolean autoMerge) {
		this.autoMerge = autoMerge;
	}

	public void setBannedUserGroups(String bannedUserGroups) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		this.bannedUserGroups = bannedUserGroups;
		if (bannedUserGroups != null)
			this.bannedUserGroupsList = StringUtil.toList(bannedUserGroups);
		else
			this.bannedUserGroupsList.clear();
	}

	public void setCurrent() {
		name = null;
	}

	public void setIsTemplate(boolean isTemplate) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		this.isTemplate = isTemplate;
	}

	public void setMergeHostToGroupMapping(String autoMergeHostToGroupMapping) {
		this.mergeHostToGroupMapping = autoMergeHostToGroupMapping;
	}

	public void setMergeURIs(String mergeURIs) {
		this.mergeURIs = mergeURIs;
	}

	public void setMetaOnly(boolean isMetaOnly) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		this.isMetaOnly = isMetaOnly;
	}
	
	public void setName(String name) {
		this.name = name;
	}
	
	public void setReadOnly() {
		this.readOnly = true;
	}

	public void setSource(String source) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		this.source = source;
	}

	public void setSslCAFiles(String sslCAFiles) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		synchronized(System.getProperties()) {
			this.sslCAFiles = sslCAFiles;
			if (sslCAFiles != null && sslCAFiles.trim().length()>0)
				System.setProperty("sslCAFiles", sslCAFiles);
		}
	}

	public synchronized void setSslCertfile(String sslCertfile) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		synchronized(System.getProperties()) {
			this.sslCertfile = sslCertfile;
			if (sslCertfile != null && sslCertfile.trim().length()>0)
				System.setProperty("sslCertfile", sslCertfile);
		}
	}
	
    public synchronized void setSslKey(String sslKey) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		synchronized(System.getProperties()) {
			this.sslKey = sslKey;
			if (sslKey != null && sslKey.trim().length()>0)
				System.setProperty("sslKey", sslKey);
		}
	}
     
    public synchronized void setSslKeyPasswd(String sslKeyPasswd) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		synchronized(System.getProperties()) {
			this.sslKeyPasswd = sslKeyPasswd;
			if (sslKeyPasswd != null && sslKeyPasswd.trim().length()>0)
				System.setProperty("sslKeyPasswd", sslKeyPasswd);
		}
	}    

	public void setVersion(String version) {
		if (readOnly)
			throw new RuntimeException("Configuration is read only");

		this.version = version;
		versionWasSet = true;
	}

	public String toXml() {
		StringWriter writer = new StringWriter();
		try {
			write(writer);
		} catch (IOException e) {
			throw new RuntimeException(e.getMessage());
		}
		return writer.toString();
	}

	public synchronized void updateBannedUserGroupMembers() throws Exception {
        if (userGroups.size()>0) {
    		failedGroups.delete(0, failedGroups.length());
        	adminLog.info("Updating user group users for all " + userGroups.size() + " banned user groups");
	        Iterator<String> iter = bannedUserGroupsList.iterator();
	        while (iter.hasNext()) {
	        	String userGroupName = iter.next();
	            UserGroup userGroup = getUserGroup(userGroupName);
	            if (userGroup == null) {
	            	String message = "User group '" + userGroupName + "' does not exist";
	            	adminLog.error(message);
	            	throw new Exception(message);
	            }
	    		while (updaterInstanceCount >= maxUpdaterThreads)
	    			Thread.sleep(1000);
	   			new UpdateThread(userGroup, true).run();
	        }
	        if (failedGroups.length() == 0) {
	        	adminLog.info("Finished updating banned users for all user groups");
	        }
	        else {
	        	adminLog.warn("Some banned user groups weren't updated correctly");
	            throw new RuntimeException("Some banned user groups weren't updated correctly" + failedGroups.toString());
	        }
        }
    }

	public synchronized void updateUserGroupMembers() {
        Collection<UserGroup> groups = getUserGroups().values();
        if (groups.size()>0) {
    		failedGroups.delete(0, failedGroups.length());
	        adminLog.info("Updating user group users for all " + groups.size() + " user groups");
	        Iterator<UserGroup> iter = groups.iterator();
	        while (iter.hasNext()) {
	            UserGroup userGroup = (UserGroup) iter.next();
	    		while (updaterInstanceCount >= maxUpdaterThreads) {
	    			try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						log.error("Sleep interrupted during usergroup update");
					}
	    		}
	   			new UpdateThread(userGroup, true).run();
	        }
	        if (failedGroups.length() == 0) {
	        	adminLog.info("Finished updating users for all user groups");
	        }
	        else {
	        	adminLog.warn("Some user groups weren't updated correctly");
	            throw new RuntimeException("Some user groups weren't updated correctly" + failedGroups.toString());
	        }
        }
    }

	public void write(Writer out) throws IOException {
		out.write("<?xml version='1.0' encoding='UTF-8'?>\n\n");

		out.write("<gums\n" +
				"\tversion='"+version+"'\n" +
				"\tallowGridmapFiles='"+isAllowGridmapFiles()+"'\n" +
				"\tbannedUserGroups='"+(getBannedUserGroups()!=null?getBannedUserGroups():"")+"'\n" +
				"\tsslCAFiles='"+(getSslCAFiles()!=null?getSslCAFiles():"")+"'\n" +
				"\tsslCertfile='"+(getSslCertfile()!=null?getSslCertfile():"")+"'\n" +
				"\tsslKey='"+(getSslKey()!=null?getSslKey():"")+"'\n" +
				"\tautoMerge='"+isAutoMerge()+"'\n" +
				"\tsslKeyPasswd='"+(getSslKeyPasswd()!=null?getSslKeyPasswd():"")+"'\n" +
				"\tmergeHostToGroupMapping='"+(getMergeHostToGroupMapping()!=null?getMergeHostToGroupMapping():"")+"'\n" +
				"\tmergeURIs='"+(getMergeURIs()!=null?getMergeURIs():"")+"'>\n\n");

		// Write User Groups
		if( getUserGroups().size()>0 ) {
			out.write("\t<userGroups>\n\n");
			Iterator<UserGroup> it = getUserGroups().values().iterator();
			while( it.hasNext() ) {
				UserGroup userGroup = it.next();
				out.write( userGroup.toXML() );
			}
			out.write("\t</userGroups>\n\n");
		}                

		// Write Account Mappers
		if( getAccountMappers().size()>0 ) {
			out.write("\t<accountMappers>\n\n");
			Iterator<AccountMapper> it = getAccountMappers().values().iterator();
			while( it.hasNext() ) {
				AccountMapper accountMapper = it.next();
				out.write( accountMapper.toXML() );
			}
			out.write("\t</accountMappers>\n\n");
		}             

		// Write Group To Account Mappings
		if( getGroupToAccountMappings().size()>0 ) {
			out.write("\t<groupToAccountMappings>\n\n");
			Iterator<GroupToAccountMapping> it = getGroupToAccountMappings().values().iterator();
			while( it.hasNext() ) {
				GroupToAccountMapping groupToAccountMapping = it.next();
				out.write( groupToAccountMapping.toXML() );
			}
			out.write("\t</groupToAccountMappings>\n\n");
		}                

		// Write Host To Group Mappings
		if( getHostToGroupMappings().size()>0 ) {
			out.write("\t<hostToGroupMappings>\n\n");
			Iterator<HostToGroupMapping> it = getHostToGroupMappings().iterator();
			while( it.hasNext() ) {
				HostToGroupMapping hostToGroupMapping = it.next();
				out.write( hostToGroupMapping.toXML() );
			}
			out.write("\t</hostToGroupMappings>\n\n");
		}                

		out.write("</gums>");   	
	}
	
	@Id
	@GeneratedValue(strategy = GenerationType.AUTO)
	@SuppressWarnings("unused")
	private int getId() {
		return id;
	}

	// Required by Hibernate
	@SuppressWarnings("unused")
	private void setAccountMappers(Map<String, AccountMapper> accountMappers) {
		this.accountMappers = accountMappers;
	}
	
	// Required by Hibernate
	@SuppressWarnings("unused")
	private void setGroupToAccountMappings(
			Map<String, GroupToAccountMapping> groupToAccountMappings) {
		this.groupToAccountMappings = groupToAccountMappings;
	}

	// Required by Hibernate
	@SuppressWarnings("unused")
	private void setHostToGroupMappings(List<HostToGroupMapping> hostToGroupMappings) {
		this.hostToGroupMappings = hostToGroupMappings;
	}

	@SuppressWarnings("unused")
	private void setId(int id) {
		this.id = id;
	}

	// Required by Hibernate
	@SuppressWarnings("unused")
	private void setUserGroups(Map<String, UserGroup> userGroups) {
		this.userGroups = userGroups;
	}
	
	protected void finalize() throws Throwable {
		super.finalize();
		instanceCount--;
	}

}
