/*
 * Group2AccountMapping.java
 *
 * Created on May 24, 2004, 2:36 PM
 */

package gov.bnl.gums.groupToAccount;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.DiscriminatorColumn;
import javax.persistence.DiscriminatorType;
import javax.persistence.Entity;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.Transient;

import gov.bnl.gums.configuration.ConfigElement;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.util.StringUtil;

/** This class defines which mapping policy should be used for the given group.
 * It tells that a given user group (all the members of the 'usatlas'
 * group in the 'atlas' LDAP vo) should be mapped using a given mapping policy
 * (a composite made of a NISAccountMapper and GroupAccountMapper, meaning
 * each user should be mapped to his account if exists, otherwise the generic
 * group account should be used).
 * <p>
 * This class is used by the Hostname mapping to hold a series of group/accountmapping
 * pairs.
 * 
 * Two ways to build the userGroups and accountMappers - one with the add/remove methods, and the other with
 * set and get methods, which Hibernate will use.
 *
 * @author Gabriele Carcassi, Jay Packard
 */
@Entity
@Inheritance(strategy=InheritanceType.SINGLE_TABLE)
@DiscriminatorColumn(
    name="type",
    discriminatorType=DiscriminatorType.STRING
)
public class GroupToAccountMapping extends ConfigElement {
    protected List<String> userGroupsList = new ArrayList<String>();
    protected List<String> accountMappersList = new ArrayList<String>();
	
	// persistent variables
    protected String accountingVo;
    protected String accountingVoSubgroup;
    protected String userGroups;
    protected String accountMappers;
    
	/**
	 * Creates a GroupToAccountMapping object. This empty constructor is needed by the XML Digestor.
	 */
	public GroupToAccountMapping() {
	}    
	
	/**
	 * Creates a GroupToAccountMapping with a configuration and a name.
	 * 
	 * @param configuration
	 * @param name
	 */
	public GroupToAccountMapping(Configuration configuration, String name) {
		super(configuration, name);
	}

	/**
     * Setter for property mapper.
     * 
     * @param mapper New value of property mapper.
     */
    public void addAccountMapper(int index, String accountMapper) {
    	if (accountMappers!=null)
    		throw new RuntimeException("Cannot call removeAccountMapper when setAccountMappers has been called");
		accountMappersList.add(index, accountMapper);
    }
    
	/**
     * Setter for property mapper.
     * 
     * @param mapper New value of property mapper.
     */
    public void addAccountMapper(String accountMapper) {
    	if (accountMappers!=null)
    		throw new RuntimeException("Cannot call removeAccountMapper when setAccountMappers has been called");
		accountMappersList.add(accountMapper);
    }

	/**
     * Setter for property mapper.
     * 
     * @param mapper New value of property mapper.
     */
    public void addUserGroup(int index, String userGroup) {
    	if (userGroups!=null)
    		throw new RuntimeException("Cannot call removeUserGroup when setUserGroups has been called");
    	userGroupsList.add(index, userGroup);
    }
    
	/**
     * Setter for property group.
     * 
     * @param group New value of property group.
     */
    public void addUserGroup(String userGroup) {
    	if (userGroups!=null)
    		throw new RuntimeException("Cannot call removeUserGroup when setUserGroups has been called");
		userGroupsList.add(userGroup);
    }
    
    /**
     * Getter for property accountingVo.
     * 
     * @return Value of property accountingVo.
     */
    public String getAccountingVo() {
        return this.accountingVo;
    }    
    
    /**
     * Getter for property accountingVoSubgroup.
     * 
     * @return Value of property accountingVoSubgroup.
     */
    public String getAccountingVoSubgroup()  {
        return this.accountingVoSubgroup;
    }    
    
    /**
     * Getter for property mapper.
     * 
     * @return Value of property mapper.
     */
    public String getAccountMappers() {
    	// Prefer string object since Hibernate requires the getter return the actual object set
    	if (accountMappers!=null)
    		return accountMappers;
    	else
    		return StringUtil.toCommaSeparatedList(accountMappersList);
    } 
    
    /**
     * Getter for property group.
     * 
     * @return Value of property group.
     */
    public String getUserGroups() {
    	// Prefer string object since Hibernate requires the getter return the actual object set
    	if (userGroups!=null)
    		return userGroups;
    	else
    		return StringUtil.toCommaSeparatedList(userGroupsList);
    }
    
    public boolean removeAccountMapper(int index) {
    	if (accountMappers!=null)
    		throw new RuntimeException("Cannot call removeAccountMapper when setAccountMappers has been called");
		return accountMappersList.remove(index)!=null;
    }
    
    public boolean removeAccountMapper(String accountMapperName) {
    	if (userGroups!=null)
    		throw new RuntimeException("Cannot call removeAccountMapper when setAccountMappers has been called");
		return accountMappersList.remove(accountMapperName);
    }
    
    public boolean removeUserGroup(int index) {
    	if (userGroups!=null)
    		throw new RuntimeException("Cannot call removeUserGroup when setUserGroups has been called");
		return userGroupsList.remove(index)!=null;
    }
    
    public boolean removeUserGroup(String userGroupName) {
    	if (userGroups!=null)
    		throw new RuntimeException("Cannot call removeUserGroup when setUserGroups has been called");
		return userGroupsList.remove(userGroupName);
    }
    
    /**
     * Setter for property accountingVo.
     * @param accountingVo New value of property y.
     */
    @ConfigFieldAnnotation(label="Accounting VO", help="to be included in OSG-user-VO-map")
    public void setAccountingVo(String accountingVo) {
        this.accountingVo = accountingVo;
    }

    /**
     * Setter for property accountingVo.
     * 
     * @param accountingVo New value of property accountingVo.
     */
    @ConfigFieldAnnotation(label="Accounting VO Subgroup", help="to be included in OSG-user-VO-map")
    public void setAccountingVoSubgroup(java.lang.String accountingVoSubgroup)  {
        this.accountingVoSubgroup = accountingVoSubgroup;
    }
    
    @ConfigFieldAnnotation (label="Account Mapper(s)", help="map using first successful account mapper", orderedList=true)
    public void setAccountMappers(String accountMappers) {
    	this.accountMappers = accountMappers;
	}
    
    @ConfigFieldAnnotation(label="User Group(s)", help="validate membership from first successful user group", orderedList=true)
    public void setUserGroups(String userGroups) {
		this.userGroups = userGroups;
	}
    
}
