/*
 * AbstractHostGroup.java
 *
 * Created on May 10, 2005, 3:44 PM
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package gov.bnl.gums.hostToGroup;

import gov.bnl.gums.configuration.ConfigElement;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.util.StringUtil;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.DiscriminatorColumn;
import javax.persistence.DiscriminatorType;
import javax.persistence.Entity;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;

/** It defines a a group of hosts that will be using the same mappings for user
 * authorization. An object of this class links a series of hostnames to a list
 * of group mappings.
 * <p>
 * This class does not return the list of hosts that will be affected since 
 * the group of hosts can be defined by a rule, without knowledge of which
 * systems exist. For example, all the machines in the 130.199.*.* subnet,
 * or all machines in the usatlas.bnl.gov subdomain.
 *
 * * Two ways to build the groupToAccountMappers - one with the add/remove methods, and the other with
 * set and get methods, which Hibernate will use.
 *
 * @author Gabriele Carcassi, Jay Packard
 */
@Entity
@Inheritance(strategy=InheritanceType.SINGLE_TABLE)
@DiscriminatorColumn(
    name="type",
    discriminatorType=DiscriminatorType.STRING
)
public abstract class HostToGroupMapping extends ConfigElement {
    protected List<String> groupToAccountMappingsList = new ArrayList<String>();
	
	//persistent variables
	String groupToAccountMappings;
	
	/**
	 * Create a new HostToGroupMapping object.
	 * 
	 * This empty constructor needed by XML Digestor
	 */
    public HostToGroupMapping() {
    }
    
	/**
	 * Create a new HostToGroupMapping object with a configuration and a name.
	 * 
	 * @param configuration
	 */
    public HostToGroupMapping(Configuration configuration, String name) {
    	super(configuration, name);
    }

    /** Changes the list of group mapping associated with this mapping.
     *
     * @param groupMapper A list of GroupMapper objects.
     */
    public void addGroupToAccountMapping(String groupToAccountMapping) {
    	if (groupToAccountMappings!=null)
    		throw new RuntimeException("Cannot call addGroupToAccountMapping when setGroupToAccountMappings has been called");
        this.groupToAccountMappingsList.add(groupToAccountMapping);
    }
    
    /** Returns the list of group mapping associated with this mapping.
     * @return A list of GroupMapper objects.
     */
    public String getGroupToAccountMappings() {
    	if (groupToAccountMappings!=null)
    		return groupToAccountMappings;
    	else
    		return StringUtil.toCommaSeparatedList(groupToAccountMappingsList);
    }
    
    /**
     * Determines if hostname is matched within this host to group mapping.
     * 
     * @param hostname
     * @return true if hostname matches
     */
    public abstract boolean isInGroup(String hostname);
    
    /**
     * @param name
     * @return
     */
    public boolean removeGroupToAccountMapping(String name) {
    	if (groupToAccountMappings!=null)
    		throw new RuntimeException("Cannot call removeGroupToAccountMapping when setGroupToAccountMappings has been called");
    	return groupToAccountMappingsList.remove(name);
    }

    @ConfigFieldAnnotation(label="Group To Account Mapping(s)", help="return account from first successful mapping", orderedList=true)
	public void setGroupToAccountMappings(String groupToAccountMappers) {
    	this.groupToAccountMappings = groupToAccountMappers;
	}
}
