/*
 * ManualUserGroup.java
 *
 * Created on May 25, 2004, 4:48 PM
 */

package gov.bnl.gums.userGroup;

import java.util.Set;

import javax.persistence.Entity;
import javax.persistence.Transient;

import org.apache.log4j.Logger;

import gov.bnl.gums.GridUser;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.util.URIToolkit;


/** 
 * A user group that is defined by a list of users stored in some way, allowing
 * to add and remove users. The persistance layer is implemented through an
 * interface, allowing different storage frameworks (i.e. database, LDAP, file)
 * <p>
 * This class will provide also configurable data caching.
 *
 * @author Gabriele Carcassi, Jay Packard
 */
@Entity
public class ManualUserGroup extends UserGroup {
    private Logger log = Logger.getLogger(ManualUserGroup.class);
    
    // persistent variables
	protected String membersUri;
	protected String nonMembersUri;
    
    public ManualUserGroup() {
    	super();
    }
    
	public ManualUserGroup(Configuration configuration, String name) {
		super(configuration, name);
	}

	public void clearCachedItems() {
	}
	
    public String getMembersUri() {
    	return membersUri;
    }

    public String getNonMembersUri() {
    	return nonMembersUri;
    }
       
	@ConfigFieldAnnotation(label="Members URI", example="file:///etc/grid-security/banned_users", help="optional")
    public void setMembersUri(String membersUri) {
    	this.membersUri = membersUri;
    }

	@ConfigFieldAnnotation(label="Non-members URI", example="file:///etc/grid-security/unbanned_users", help="optional")
    public void setNonMembersUri(String nonMembersUri) {
    	this.nonMembersUri = nonMembersUri;
    }

	public void addMember(GridUser gridUser) {
    	try {
    		synchronized(members) {
    			gridUser.setUserGroup(this);
    			members.add(gridUser);
    		}
	        refreshPatternList();
    	}
    	catch(Exception e) {
    		log.error(e);
    	}
    }
    
	public boolean removeMember(GridUser gridUser) {
    	boolean value = false;
    	try {
    		synchronized(members) {
    			value = members.remove(gridUser);
    		}
    		refreshPatternList();
		}
		catch(Exception e) {
			log.error(e);
		}
        return value;
    }
	
    protected Set<GridUser> retrieveMembers() {
    	// add to or remove from members and nonmembers URI
    	synchronized(members) {
	    	try {
		    	if (membersUri!=null && membersUri.length()>0) {
		    		String str = URIToolkit.parseUri(membersUri, null);
		    		String[] lines = str.split("\n");
		    		for (String line: lines) {
		    			if (line.length()==0)
		    				continue;
		    			String[] elements = line.split(",");
		    			GridUser user = new GridUser();
		    			user.initialize(elements);
		    			if (!members.contains(user)) {
		    				log.info("Added user "+user+" to "+getName()+" user group");
		    				members.add(user);
		    			}
		    		}
		    	}
		    } catch (Throwable e) {
		    	String message = "Couldn't retrieve users from URI " + membersUri + ": ";
		        log.error(message, e);
		        throw new RuntimeException(message + ": " + e.getMessage());
		    }
		    
		    try {
		    	if (nonMembersUri!=null && nonMembersUri.length()>0) {
		    		String str = URIToolkit.parseUri(nonMembersUri, null);
		    		String[] lines = str.split("\n");
		    		for (String line: lines) {
		    			if (line.length()==0)
		    				continue;
		    			String[] elements = line.split(",");
		    			GridUser user = new GridUser();
		    			user.initialize(elements);
		    			if (members.contains(user)) {
		    				log.info("Removed user "+user+" from "+getName()+" user group");
		    				members.remove(user);
		    			}
		    		}
		    	}
		    } catch (Throwable e) {
		    	String message = "Couldn't retrieve users from URI " + nonMembersUri + ": ";
		        log.error(message, e);
		        throw new RuntimeException(message + ": " + e.getMessage());
		    }
		}
		return members;
    }
}
