/*
 * VOMSUserGroupManager.java
 *
 * Created on May 25, 2004, 10:20 AM
 */

package gov.bnl.gums.userGroup;

import gov.bnl.gums.FQAN;
import gov.bnl.gums.GridUser;
import gov.bnl.gums.configuration.Configuration;
import gov.bnl.gums.util.StringUtil;

import java.net.URL;
import java.util.*;

import javax.persistence.Entity;
import javax.persistence.Transient;

import org.apache.log4j.Logger; 

import org.glite.security.voms.*;


/** A group of users residing on a VOMS vo database. This class is able to 
 * import a list of users from a VOMS server. It will store to a local
 * medium through the UserGroupDB interface. It also manages the caching from
 * the local database.
 * <p>
 * The authentication is done through the proxy, or a certificate/key/password
 * combination. The parameters are to be set externally as system properties.
 * The proxy can be set through "gridProxyFile" property. Other properties
 * are "sslCertfile", "sslKey", "sslKeyPasswd" and "sslCAFiles". More documentation
 * can be found in the documentation of the edg trustmanager  
 *
 * @author Gabriele Carcassi, Jay Packard
 */
@Entity
public class VOMSUserGroup extends UserGroup {
	static protected final boolean defaultAcceptProxyWithoutFQAN = true;
	static protected String[] matchFQANTypes = {"ignore", "exact", "vorole", "role", "vogroup", "vo"};
	static protected final int defaultMatchFQANIndex = 0;
	protected String matchFQAN = matchFQANTypes[defaultMatchFQANIndex];
	static private Logger log = Logger.getLogger(VOMSUserGroup.class);

    static public Set<String> getMatchFQANTypes() {
		Set<String> retList = new TreeSet<String>();
		for(int i=0; i<matchFQANTypes.length; i++)
			retList.add(matchFQANTypes[i]);
		return retList;
	}
    
    // persistent variables
    protected boolean acceptProxyWithoutFQAN = defaultAcceptProxyWithoutFQAN;
    protected String fqan = null;
    protected String role;
    protected String voGroup;
    protected String vomsServerUrls;
    
    static {
    	System.setProperty("axis.socketSecureFactory", "org.glite.security.trustmanager.axis.AXISSocketFactory");
    }
    
    public VOMSUserGroup() {
    	super();
    	hasCache = true;
    }  

	public VOMSUserGroup(Configuration configuration, String name) {
		super(configuration, name);
    	hasCache = true;
	}
	
    /**
     * The scheme according to which the FQAN will be matched.
     * <p>
     * Possible values are:
     * <ul>
     *   <li>exact (default) - role, group, and vo have to match. </li>
     *   <li>vorole - role and vo have to match.</li>
     *   <li>role - rolehas to match.</li>
     *   <li>group, vogroup - group and vo have to match.</li>
     *   <li>vo - vo has to match.</li>
     *   <li>ignore - no matching.</li>
     * </ul>
     * @return matching type as String.
     */
    public String getMatchFQAN() {
   		return matchFQAN;
    }
    
    /**
     * Changes the role.
     * @return The role name in the VOMS server (i.e. myrole), or "" for no role
     */
    public String getRole() {
        return this.role;
    }
    
    /**
     * Returns the VO group.
     * @return The group in the VOMS (i.e. /atlas/usatlas)
     */
    public String getVoGroup() {
        return this.voGroup;
    }

    public String getVomsServerUrls() {
    	return vomsServerUrls;
	}

    /**
     * True if non-VOMS will be accepted. If true, all non-VOMS proxies with a matchin
     * DN will be matched. VOMS proxies won't be affected by the use of this property.
     * @return True if group will accept non-VOMS proxies
     */
    public boolean isAcceptProxyWithoutFQAN() {
        return this.acceptProxyWithoutFQAN;
    }
    
    /**
     * Convenience function for "ignore".equals(getmatchFQAN())
     * @return False if FQAN is used during the match
     */
   public boolean isIgnoreFQAN() {
        return matchFQAN.equals(matchFQANTypes[0]);
    }

    public boolean isMember(GridUser gridUser) {
    	if (gridUser.getFqanObj() == null) {
            // If the user comes in without FQAN and we don't accept proxies without fqan,
            // kick him out right away
	        if (!isAcceptProxyWithoutFQAN())
	            return false;
	        
	        // If the user comes in without FQAN and we accept proxies without it,
	        // we simply check whether the DN is in the database
	        else
        		return isMember(new GridUser(gridUser.getDn(), fqan));
    	}

        // We now know we don't have user.getFqanObj()==null

        // If we have vorole match, entire fqan has to be the same
        if ("exact".equals(getMatchFQAN())) {
            if (!gridUser.getFqan().equals(fqan))
                return false;
        }
        
        // If we have a vo-role match, vo and role has to be the same
        if ("vorole".equals(getMatchFQAN()) && gridUser.getFqanObj().getVo()!=null && gridUser.getFqanObj().getRole()!=null) {
        	FQAN theFQAN = new FQAN(fqan);
            if (!gridUser.getFqanObj().getVo().equals(theFQAN.getVo()) && !gridUser.getFqanObj().getRole().equals(theFQAN.getRole()))
                return false;
        }
        
        // If we have a role match, role has to be the same
        if ("role".equals(getMatchFQAN()) && gridUser.getFqanObj().getRole()!=null) {
        	FQAN theFQAN = new FQAN(fqan);
            if (!gridUser.getFqanObj().getRole().equals(theFQAN.getRole()))
                return false;
        }
        
        // If we match the group, we make sure the VO starts with the group
        if ("group".equals(getMatchFQAN()) || "vogroup".equals(getMatchFQAN())) {
            if (!gridUser.getFqanObj().toString().startsWith(voGroup))
                return false;
        }

        // If we match the vo, we check the vo is the same
        if ("vo".equals(getMatchFQAN()) && gridUser.getFqanObj().getVo()!=null) {
            FQAN theFQAN = new FQAN(fqan);
            if (!gridUser.getFqanObj().getVo().equals(theFQAN.getVo()))
                return false;
        }
        
        // FQAN matches, let's look up if the DN is in the db
        // If not, he's kicked out
        return super.isMember(new GridUser(gridUser.getDn(), fqan));
    }
    
    /**
     * Changes the way non-VOMS proxies are handled.
     * @param acceptProxyWithoutFQAN True if group will accept non-VOMS proxies
     */
    @ConfigFieldAnnotation(label="Accept non-VOMS certificates")
    public void setAcceptProxyWithoutFQAN(boolean acceptProxyWithoutFQAN) {
        this.acceptProxyWithoutFQAN = acceptProxyWithoutFQAN;
    }

    @Transient
    public void setIgnoreFQAN(boolean ignoreFQAN) {
    	matchFQAN = matchFQANTypes[0];
    }
    
    /**
     * Changes the scheme according to which the FQAN will be matched. See
     * getMatchFQAN for more details.
     * @param matchFQAN One of the following:  "exact, "vorole, "role", "vogroup", "vo", "ignore". (also "group" for backwards compat.)
     */
    @ConfigFieldAnnotation(label="Match VOMS certificate's FQAN as", choices="ignore, exact, vorole, role, vogroup, vo")
    public void setMatchFQAN(String matchFQAN) {
    	for(int i=0; i<matchFQANTypes.length; i++) {
	    	if (matchFQAN.equals("group"))
	    		matchFQAN = "vogroup";
	    	if (matchFQAN.equals(""))
	    		matchFQAN = "exact";
    		if ( matchFQANTypes[i].equalsIgnoreCase(matchFQAN) ) {
    	    	this.matchFQAN = matchFQAN;
    			return;
    		}
    	}
		throw new RuntimeException("Invalid match FQAN string: "+matchFQAN);
    }
    
    /**
     * Changes the role.
     * @param role The role in the VOMS (i.e.production)
     */
    @ConfigFieldAnnotation(label="Role", order=7) 
    public void setRole(String role) {
        this.role = role;
        prepareFQAN();
    }
    
	/**
     * Changes the VO group.
     * @param voGroup The group in the VOMS (i.e. /atlas/usatlas)
     */
    @ConfigFieldAnnotation(label="VO/Group", example="/atlas/usatlas", help="optional", order=6) 
    public void setVoGroup(String voGroup) {
        this.voGroup = voGroup;
        prepareFQAN();
    }
    
    @ConfigFieldAnnotation(label="VOMS Server URLs", size=64, example="https://lcg-voms.cern.ch:8443/voms/atlas/services/VOMSAdmin", order=5) 
	public void setVomsServerUrls(String vomsServerUrls) {
    	this.vomsServerUrls = vomsServerUrls;
	}

	protected void prepareFQAN() {
        if (role != null && voGroup != null)
        	fqan = voGroup + "/Role=" + role;
        else if (voGroup != null)
            fqan = voGroup;
        else
        	fqan = null;
    }
	
	protected Set<GridUser> retrieveMembers() {
    	if (getConfiguration()==null)
    		throw new RuntimeException("Configuration has not yet been set for this class");
    	
    	Set<GridUser> entries = new HashSet<GridUser>();
    	Iterator<String> it = StringUtil.toList(vomsServerUrls).iterator();
    	while (it.hasNext()) {
    		String vomsServerUrl = it.next();
	        try {
	            VOMSAdminServiceLocator locator = new VOMSAdminServiceLocator();
	            log.info("Connecting to VOMS admin at " + vomsServerUrl);
	            VOMSAdmin voms = locator.getVOMSAdmin(new URL( vomsServerUrl ));
	        	org.glite.security.voms.User[] users = null;
	            if (role == null) {
	                users = voms.listMembers( getVoGroup() );
	            } else if(voGroup != null) {
	                users = voms.listUsersWithRole( getVoGroup(), "Role=" + getRole());
	            }
	            if (users.length > 0) {
	                log.info("Retrieved " + users.length + " users. First is: '" + users[0].getDN() + "'");
	            } else {
	                log.info("Retrieved no users.");
	            }
	            for (int n=0; n < users.length; n++) {
	            	GridUser gridUser = new GridUser(users[n].getDN(), fqan, users[n].getMail());
	                entries.add(gridUser);
	            }
	            break;
	        } catch (Throwable e) {
	        	String message = "Couldn't retrieve users: ";
	        	if (!it.hasNext()) {
	        		log.error(message, e);
		            throw new RuntimeException(message + e.getMessage());
	        	}
	        	else {
	        		log.warn(message, e);
	        	}
	        }
    	}
    	return entries;
    }
}
